"""
Integration tests for the docker_container states
"""

import logging
import os
import random
import shutil
import subprocess

import attr
import pytest
from saltfactories.utils import random_string

import salt.utils.files
import salt.utils.network
import salt.utils.path
from salt._compat import ipaddress
from salt.exceptions import CommandExecutionError
from salt.modules.config import DEFAULTS as _config_defaults
from tests.support.runtests import RUNTIME_VARS

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_on_freebsd(reason="No Docker on FreeBSD available"),
    pytest.mark.skip_if_binaries_missing("busybox", reason="Busybox not installed"),
    pytest.mark.skip_if_binaries_missing(
        "docker", "dockerd", reason="Docker not installed"
    ),
    pytest.mark.timeout_unless_on_windows(240),
]

IPV6_ENABLED = bool(salt.utils.network.ip_addrs6(include_loopback=True))


@attr.s(slots=True)
class Network:

    docker_exec_mod = attr.ib(repr=False)
    subnet = attr.ib()
    name = attr.ib()
    net = attr.ib(init=False)
    gateway = attr.ib()
    enable_ipv6 = attr.ib()
    ip_arg = attr.ib(init=False)
    _rand_indexes = attr.ib(init=False)

    @name.default
    def _name(self):
        return random_string("salt-network-", uppercase=False)

    @net.default
    def _net(self):
        return ipaddress.ip_network(self.subnet)

    @_rand_indexes.default
    def __rand_indexes(self):  # pylint: disable=unused-private-member
        return random.sample(
            range(2, self.net.num_addresses - 1), self.net.num_addresses - 3
        )

    @ip_arg.default
    def _ip_arg(self):
        return f"ipv{self.net.version}_address"

    @enable_ipv6.default
    def _enable_ipv6(self):
        return self.net.version == 6

    @staticmethod
    def arg_map(arg_name):
        if arg_name == "ipv4_address":
            return "IPv4Address"
        if arg_name == "ipv6_address":
            return "IPv6Address"
        if arg_name == "links":
            return "Links"
        if arg_name == "aliases":
            return "Aliases"

    @property
    def compressed_subnet(self):
        if self.net:
            return self.net.compressed

    @gateway.default
    def _gateway(self):
        return self.net[1].compressed

    def __getitem__(self, index):
        try:
            return self.net[self._rand_indexes[index]].compressed
        except (TypeError, AttributeError):
            raise ValueError(
                "Indexing not supported for networks without a custom subnet"
            )

    def __enter__(self):
        self.docker_exec_mod.create_network(
            self.name,
            subnet=self.compressed_subnet,
            gateway=self.gateway,
            enable_ipv6=self.enable_ipv6,
        )
        return self

    def __exit__(self, *_):
        try:
            self.docker_exec_mod.disconnect_all_containers_from_network(self.name)
        except CommandExecutionError as exc:
            if "404" not in str(exc):
                raise
        self.docker_exec_mod.remove_network(self.name)


@attr.s(slots=True)
class Networks:
    docker_exec_mod = attr.ib(repr=False)
    nets = attr.ib(default=attr.Factory(list))

    def add(self, subnet):
        network = Network(docker_exec_mod=self.docker_exec_mod, subnet=subnet)
        self.nets.append(network)
        return network

    def __enter__(self):
        for net in self.nets:
            net.__enter__()
        return self

    def __exit__(self, *_):
        for net in self.nets:
            net.__exit__()


@pytest.fixture
def network(modules):
    def _network(**kwargs):
        return Network(docker_exec_mod=modules.docker, **kwargs)

    return _network


@pytest.fixture
def networks(modules):
    return Networks(docker_exec_mod=modules.docker)


@pytest.fixture
def container_name(modules):
    """
    Generate a randomized name for a container and clean it up afterward
    """
    _container_name = random_string("salt-docker-container-test-", uppercase=False)
    try:
        yield _container_name
    finally:
        try:
            modules.docker.rm(_container_name, force=True)
        except CommandExecutionError as exc:
            if "No such container" not in str(exc):
                raise


@pytest.fixture
def docker_container(states):
    return states.docker_container


@pytest.fixture(scope="module")
def image(grains, tmp_path_factory):
    if grains["os"] == "VMware Photon OS" and grains["osmajorrelease"] == 5:
        pytest.skip(f"Temporary skip on {grains['osfinger']}")

    if not salt.utils.path.which("docker"):
        # Somehow the above skip_if_binaries_missing marker for docker
        # only get's evaluated after this fixture?!?
        pytest.skip("The `docker` binary is not available")
    container_build_dir = tmp_path_factory.mktemp("busybox")
    image_name = random_string("salt-busybox-", uppercase=False)

    script_path = os.path.join(RUNTIME_VARS.BASE_FILES, "mkimage-busybox-static")
    cmd = [script_path, str(container_build_dir), image_name]
    log.debug("Running '%s' to build busybox image", " ".join(cmd))
    process = subprocess.run(
        cmd,
        close_fds=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        check=False,
    )

    log.debug("Output from mkimge-busybox-static:\n%s", process.stdout)
    if process.returncode != 0:
        raise Exception(
            "Failed to build image. Output from mkimge-busybox-static:\n{}".format(
                process.stdout
            )
        )

    shutil.rmtree(str(container_build_dir))

    try:
        yield image_name
    finally:
        cmd = ["docker", "rmi", "--force", image_name]
        log.debug("Running '%s' to destroy busybox image", " ".join(cmd))
        process = subprocess.run(
            cmd,
            close_fds=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            check=False,
        )
        log.debug("Output from %s:\n%s", " ".join(cmd), process.stdout)

        if process.returncode != 0:
            raise Exception("Failed to destroy image")


@pytest.mark.slow_test
def test_running_with_no_predefined_volume(
    docker_container, container_name, image, tmp_path, modules
):
    """
    This tests that a container created using the docker_container.running
    state, with binds defined, will also create the corresponding volumes
    if they aren't pre-defined in the image.
    """
    ret = docker_container.running(
        name=container_name,
        image=image,
        binds=f"{tmp_path}:/foo",
        shutdown_timeout=1,
    )
    assert ret.result is True
    # Now check to ensure that the container has volumes to match the
    # binds that we used when creating it.
    ret = modules.docker.inspect_container(container_name)
    assert "/foo" in ret["Config"]["Volumes"]


@pytest.mark.slow_test
def test_running_with_no_predefined_ports(
    docker_container, container_name, image, modules, subtests
):
    """
    This tests that a container created using the docker_container.running
    state, with port_bindings defined, will also configure the
    corresponding ports if they aren't pre-defined in the image.
    """
    ret = docker_container.running(
        name=container_name,
        image=image,
        port_bindings="14505-14506:24505-24506,2123:2123/udp,8080",
        shutdown_timeout=1,
    )
    assert ret.result is True
    # Now check to ensure that the container has ports to match the
    # port_bindings that we used when creating it.
    expected_ports = ("8080/tcp", "2123/udp")
    ret = modules.docker.inspect_container(container_name)
    for x in expected_ports:
        with subtests.test(port=x):
            assert x in ret["NetworkSettings"]["Ports"]


@pytest.mark.slow_test
def test_running_updated_image_id(docker_container, container_name, image, modules):
    """
    This tests the case of an image being changed after the container is
    created. The next time the state is run, the container should be
    replaced because the image ID is now different.
    """
    # Create and start a container
    ret = docker_container.running(name=container_name, image=image, shutdown_timeout=1)
    assert ret.result is True
    # Get the container's info
    c_info = modules.docker.inspect_container(container_name)
    c_name, c_id = (c_info[x] for x in ("Name", "Id"))
    # Alter the filesystem inside the container
    assert modules.docker.retcode(container_name, "touch /.salttest") == 0
    # Commit the changes and overwrite the test class' image
    modules.docker.commit(c_id, image)
    # Re-run the state
    ret = docker_container.running(name=container_name, image=image, shutdown_timeout=1)
    assert ret.result is True
    # Check to make sure that the container was replaced
    assert "container_id" in ret.changes
    # Check to make sure that the image is in the changes dict, since
    # it should have changed
    assert "image" in ret.changes
    # Check that the comment in the state return states that
    # container's image has changed
    assert "Container has a new image" in ret.comment


@pytest.mark.slow_test
def test_running_start_false_without_replace(
    docker_container, container_name, image, modules
):
    """
    Test that we do not start a container which is stopped, when it is not
    being replaced.
    """
    # Create a container
    ret = docker_container.running(name=container_name, image=image, shutdown_timeout=1)
    assert ret.result is True
    # Stop the container
    modules.docker.stop(container_name, force=True)
    # Re-run the state with start=False
    ret = docker_container.running(
        name=container_name,
        image=image,
        start=False,
        shutdown_timeout=1,
    )
    assert ret.result is True
    # Check to make sure that the container was not replaced
    assert "container_id" not in ret.changes
    # Check to make sure that the state is not the changes dict, since
    # it should not have changed
    assert "state" not in ret.changes


@pytest.mark.slow_test
def test_running_no_changes_hostname_network(
    docker_container, container_name, image, network
):
    """
    Test that changes are not detected when a hostname is specified for a container
    on a custom network
    """
    with network(subnet="10.247.197.96/27") as net:
        # Create a container
        kwargs = {
            "name": container_name,
            "image": image,
            "shutdown_timeout": 1,
            "network_mode": net.name,
            "networks": [net.name],
            "hostname": "foo",
        }
        ret = docker_container.running(**kwargs)
        assert ret.result is True

        ret = docker_container.running(**kwargs)
        assert ret.result is True
        # Should be no changes
        assert not ret.changes


@pytest.mark.slow_test
def test_running_start_false_with_replace(
    docker_container, container_name, image, modules
):
    """
    Test that we do start a container which was previously stopped, even
    though start=False, because the container was replaced.
    """
    # Create a container
    ret = docker_container.running(name=container_name, image=image, shutdown_timeout=1)
    assert ret.result is True
    # Stop the container
    modules.docker.stop(container_name, force=True)
    # Re-run the state with start=False but also change the command to
    # trigger the container being replaced.
    ret = docker_container.running(
        name=container_name,
        image=image,
        command="sleep 600",
        start=False,
        shutdown_timeout=1,
    )
    assert ret.result is True
    # Check to make sure that the container was not replaced
    assert "container_id" in ret.changes
    # Check to make sure that the state is not the changes dict, since
    # it should not have changed
    assert "state" not in ret.changes


@pytest.mark.slow_test
def test_running_start_true(docker_container, container_name, image, modules):
    """
    This tests that we *do* start a container that is stopped, when the
    "start" argument is set to True.
    """
    # Create a container
    ret = docker_container.running(name=container_name, image=image, shutdown_timeout=1)
    assert ret.result is True
    # Stop the container
    modules.docker.stop(container_name, force=True)
    # Re-run the state with start=True
    ret = docker_container.running(
        name=container_name,
        image=image,
        start=True,
        shutdown_timeout=1,
    )
    assert ret.result is True
    # Check to make sure that the container was not replaced
    assert "container_id" not in ret.changes
    # Check to make sure that the state is in the changes dict, since
    # it should have changed
    assert "state" in ret.changes
    # Check that the comment in the state return states that
    # container's state has changed
    assert "State changed from 'stopped' to 'running'" in ret.comment


def test_running_with_invalid_input(docker_container, container_name, image):
    """
    This tests that the input tranlation code identifies invalid input and
    includes information about that invalid argument in the state return.
    """
    # Try to create a container with invalid input
    ret = docker_container.running(
        name=container_name,
        image=image,
        ulimits="nofile:2048",
        shutdown_timeout=1,
    )
    assert ret.result is False
    # Check to make sure that the container was not created
    assert "container_id" not in ret.changes
    # Check that the error message about the invalid argument is
    # included in the comment for the state
    assert (
        "Ulimit definition 'nofile:2048' is not in the format"
        " type=soft_limit[:hard_limit]" in ret.comment
    )


def test_running_with_argument_collision(docker_container, container_name, image):
    """
    this tests that the input tranlation code identifies an argument
    collision (API args and their aliases being simultaneously used) and
    includes information about them in the state return.
    """
    # try to create a container with invalid input
    ret = docker_container.running(
        name=container_name,
        image=image,
        ulimits="nofile=2048",
        ulimit="nofile=1024:2048",
        shutdown_timeout=1,
    )
    assert ret.result is False
    # Check to make sure that the container was not created
    assert "container_id" not in ret.changes
    # Check that the error message about the collision is included in
    # the comment for the state
    assert "'ulimit' is an alias for 'ulimits'" in ret.comment


@pytest.mark.slow_test
def test_running_with_ignore_collisions(
    docker_container, container_name, image, modules
):
    """
    This tests that the input tranlation code identifies an argument
    collision (API args and their aliases being simultaneously used)
    includes information about them in the state return.
    """
    # try to create a container with invalid input
    ret = docker_container.running(
        name=container_name,
        image=image,
        ignore_collisions=True,
        ulimits="nofile=2048",
        ulimit="nofile=1024:2048",
        shutdown_timeout=1,
    )
    assert ret.result is True
    # Check to make sure that the container was created
    assert "container_id" in ret.changes
    # Check that the value from the API argument was one that was used
    # to create the container
    c_info = modules.docker.inspect_container(container_name)
    actual = c_info["HostConfig"]["Ulimits"]
    expected = [{"Name": "nofile", "Soft": 2048, "Hard": 2048}]
    assert actual == expected


@pytest.mark.slow_test
def test_running_with_removed_argument(
    docker_container, container_name, image, modules
):
    """
    This tests that removing an argument from a created container will
    be detected and result in the container being replaced.

    It also tests that we revert back to the value from the image. This
    way, when the "command" argument is removed, we confirm that we are
    reverting back to the image's command.
    """
    # Create the container
    ret = docker_container.running(
        name=container_name,
        image=image,
        command="sleep 600",
        shutdown_timeout=1,
    )
    assert ret.result is True
    # Run the state again with the "command" argument removed
    ret = docker_container.running(name=container_name, image=image, shutdown_timeout=1)
    assert ret.result is True
    # Now check to ensure that the changes include the command
    # reverting back to the image's command.
    image_info = modules.docker.inspect_image(image)
    assert (
        ret.changes["container"]["Config"]["Cmd"]["new"] == image_info["Config"]["Cmd"]
    )


@pytest.mark.slow_test
def test_running_with_port_bindings(docker_container, container_name, image, modules):
    """
    This tests that the ports which are being bound are also exposed, even
    when not explicitly configured. This test will create a container with
    only some of the ports exposed, including some which aren't even bound.
    The resulting containers exposed ports should contain all of the ports
    defined in the "ports" argument, as well as each of the ports which are
    being bound.
    """
    # Create the container
    ret = docker_container.running(
        name=container_name,
        image=image,
        command="sleep 600",
        shutdown_timeout=1,
        port_bindings=[1234, "1235-1236", "2234/udp", "2235-2236/udp"],
        ports=[1235, "2235/udp", 9999],
    )
    assert ret.result is True

    # Check the created container's port bindings and exposed ports. The
    # port bindings should only contain the ports defined in the
    # port_bindings argument, while the exposed ports should also contain
    # the extra port (9999/tcp) which was included in the ports argument.
    cinfo = modules.docker.inspect_container(container_name)
    ports = ["1234/tcp", "1235/tcp", "1236/tcp", "2234/udp", "2235/udp", "2236/udp"]
    assert sorted(cinfo["HostConfig"]["PortBindings"]) == ports
    assert sorted(cinfo["Config"]["ExposedPorts"]) == ports + ["9999/tcp"]


@pytest.mark.slow_test
def test_absent_with_stopped_container(
    docker_container, container_name, image, modules
):
    """
    This tests the docker_container.absent state on a stopped container
    """
    # Create the container
    modules.docker.create(image, name=container_name)
    # Remove the container
    ret = docker_container.absent(
        name=container_name,
    )
    assert ret.result is True
    # Check that we have a removed container ID in the changes dict
    assert "removed" in ret.changes

    # Run the state again to confirm it changes nothing
    ret = docker_container.absent(
        name=container_name,
    )
    assert ret.result is True
    # Nothing should have changed
    assert ret.changes == {}
    # Ensure that the comment field says the container does not exist
    assert ret.comment == f"Container '{container_name}' does not exist"


@pytest.mark.slow_test
def test_absent_with_running_container(docker_container, container_name, image):
    """
    This tests the docker_container.absent state and
    """
    # Create the container
    ret = docker_container.running(
        name=container_name,
        image=image,
        command="sleep 600",
        shutdown_timeout=1,
    )
    assert ret.result is True

    # Try to remove the container. This should fail because force=True
    # is needed to remove a container that is running.
    ret = docker_container.absent(
        name=container_name,
    )
    assert ret.result is False
    # Nothing should have changed
    assert ret.changes == {}
    # Ensure that the comment states that force=True is required
    assert (
        ret.comment == "Container is running, set force to True to forcibly remove it"
    )

    # Try again with force=True. This should succeed.
    ret = docker_container.absent(
        name=container_name,
        force=True,
    )
    assert ret.result is True
    # Check that we have a removed container ID in the changes dict
    assert "removed" in ret.changes
    # The comment should mention that the container was removed
    assert ret.comment == f"Forcibly removed container '{container_name}'"


@pytest.mark.slow_test
def test_running_image_name(docker_container, container_name, image, modules):
    """
    Ensure that we create the container using the image name instead of ID
    """
    ret = docker_container.running(name=container_name, image=image, shutdown_timeout=1)
    assert ret.result is True
    ret = modules.docker.inspect_container(container_name)
    assert ret["Config"]["Image"] == image


@pytest.mark.slow_test
def test_env_with_running_container(docker_container, container_name, image, modules):
    """
    docker_container.running environnment part. Testing issue 39838.
    """
    ret = docker_container.running(
        name=container_name,
        image=image,
        env="VAR1=value1,VAR2=value2,VAR3=value3",
        shutdown_timeout=1,
    )
    assert ret.result is True
    ret = modules.docker.inspect_container(container_name)
    assert "VAR1=value1" in ret["Config"]["Env"]
    assert "VAR2=value2" in ret["Config"]["Env"]
    assert "VAR3=value3" in ret["Config"]["Env"]
    ret = docker_container.running(
        name=container_name,
        image=image,
        env="VAR1=value1,VAR2=value2",
        shutdown_timeout=1,
    )
    assert ret.result is True
    ret = modules.docker.inspect_container(container_name)
    assert "VAR1=value1" in ret["Config"]["Env"]
    assert "VAR2=value2" in ret["Config"]["Env"]
    assert "VAR3=value3" not in ret["Config"]["Env"]


@pytest.mark.slow_test
def test_static_ip_one_network(
    docker_container, container_name, image, modules, network
):
    """
    Ensure that if a network is created and specified as network_mode, that is the only network, and
    the static IP is applied.
    """
    with network(subnet="10.247.197.96/27") as net:
        requested_ip = "10.247.197.100"
        kwargs = {
            "name": container_name,
            "image": image,
            "network_mode": net.name,
            "networks": [{net.name: [{"ipv4_address": requested_ip}]}],
            "shutdown_timeout": 1,
        }
        # Create a container
        ret = docker_container.running(**kwargs)
        assert ret.result is True

        inspect_result = modules.docker.inspect_container(container_name)
        connected_networks = inspect_result["NetworkSettings"]["Networks"]

        assert list(connected_networks.keys()) == [net.name]
        assert inspect_result["HostConfig"]["NetworkMode"] == net.name
        assert connected_networks[net.name]["IPAMConfig"]["IPv4Address"] == requested_ip


def _test_running(docker_container, container_name, image, modules, *nets):
    """
    DRY function for testing static IPs
    """


@pytest.mark.slow_test
@pytest.mark.parametrize(
    "subnets",
    [
        (dict(subnet="10.247.197.96/27"),),
        (dict(subnet="10.247.197.128/27"), dict(subnet="10.247.197.96/27")),
        (dict(subnet="fe3f:2180:26:1::/123"),),
        (dict(subnet="fe3f:2180:26:1::20/123"), dict(subnet="fe3f:2180:26:1::/123")),
        (dict(subnet="fe3f:2180:26:1::/123"), dict(subnet="10.247.197.96/27")),
    ],
)
def test_running_networks(
    docker_container, container_name, image, subnets, networks, modules
):

    netdefs = []
    for subnet in subnets:
        network = networks.add(**subnet)
        if network.enable_ipv6 and IPV6_ENABLED is False:
            pytest.skip("IPv6 not enabled")

        netdefs.append({network.name: [{network.ip_arg: network[0]}]})

    with networks:

        kwargs = {
            "name": container_name,
            "image": image,
            "networks": netdefs,
            "shutdown_timeout": 1,
        }
        # Create a container
        ret = docker_container.running(**kwargs)
        assert ret.result is True

        inspect_result = modules.docker.inspect_container(container_name)
        connected_networks = inspect_result["NetworkSettings"]["Networks"]

        # Check that the correct IP was set
        try:
            for net in networks.nets:
                assert (
                    connected_networks[net.name]["IPAMConfig"][net.arg_map(net.ip_arg)]
                    == net[0]
                )
        except KeyError:
            # Fail with a meaningful error
            msg = "Container does not have the expected network config for network {}".format(
                net.name
            )
            log.error(msg)
            log.error("Connected networks: %s", connected_networks)
            pytest.fail(f"{msg}. See log for more information.")

        # Check that container continued running and didn't immediately exit
        assert inspect_result["State"]["Running"]

        # Update the SLS configuration to use the second random IP so that we
        # can test updating a container's network configuration without
        # replacing the container.
        for idx, net in enumerate(networks.nets):
            kwargs["networks"][idx][net.name][0][net.ip_arg] = net[1]
        ret = docker_container.running(**kwargs)
        assert ret.result is True

        expected = {"container": {"Networks": {}}}
        for net in networks.nets:
            expected["container"]["Networks"][net.name] = {
                "IPAMConfig": {
                    "old": {net.arg_map(net.ip_arg): net[0]},
                    "new": {net.arg_map(net.ip_arg): net[1]},
                }
            }
        assert ret.changes == expected

        expected = [f"Container '{container_name}' is already configured as specified."]
        expected.extend(
            [
                f"Reconnected to network '{x.name}' with updated configuration."
                for x in sorted(networks.nets, key=lambda y: y.name)
            ]
        )
        expected = " ".join(expected)
        assert ret.comment == expected

        # Update the SLS configuration to remove the last network
        kwargs["networks"].pop(-1)
        ret = docker_container.running(**kwargs)
        assert ret.result is True

        expected = {
            "container": {
                "Networks": {
                    networks.nets[-1].name: {
                        "IPAMConfig": {
                            "old": {
                                networks.nets[-1].arg_map(
                                    networks.nets[-1].ip_arg
                                ): networks.nets[-1][1]
                            },
                            "new": None,
                        }
                    }
                }
            }
        }
        assert ret.changes == expected

        expected = (
            "Container '{}' is already configured as specified. Disconnected "
            "from network '{}'.".format(container_name, networks.nets[-1].name)
        )
        assert ret.comment == expected

        # Update the SLS configuration to add back the last network, only use
        # an automatic IP instead of static IP.
        kwargs["networks"].append(networks.nets[-1].name)
        ret = docker_container.running(**kwargs)
        assert ret.result is True

        # Get the automatic IP by inspecting the container, and use it to build
        # the expected changes.
        container_netinfo = (
            modules.docker.inspect_container(container_name)
            .get("NetworkSettings", {})
            .get("Networks", {})[networks.nets[-1].name]
        )
        autoip_keys = _config_defaults["docker.compare_container_networks"]["automatic"]
        autoip_config = {
            x: y for x, y in container_netinfo.items() if x in autoip_keys and y
        }

        expected = {"container": {"Networks": {networks.nets[-1].name: {}}}}
        for key, val in autoip_config.items():
            expected["container"]["Networks"][networks.nets[-1].name][key] = {
                "old": None,
                "new": val,
            }
        assert ret.changes == expected

        expected = (
            "Container '{}' is already configured as specified. Connected "
            "to network '{}'.".format(container_name, networks.nets[-1].name)
        )
        assert ret.comment == expected

        # Update the SLS configuration to remove the last network
        kwargs["networks"].pop(-1)
        ret = docker_container.running(**kwargs)
        assert ret.result is True

        expected = {"container": {"Networks": {networks.nets[-1].name: {}}}}
        for key, val in autoip_config.items():
            expected["container"]["Networks"][networks.nets[-1].name][key] = {
                "old": val,
                "new": None,
            }
        assert ret.changes == expected

        expected = (
            "Container '{}' is already configured as specified. Disconnected "
            "from network '{}'.".format(container_name, networks.nets[-1].name)
        )
        assert ret.comment == expected


def test_running_explicit_networks(
    docker_container, container_name, image, modules, network
):
    """
    Ensure that if we use an explicit network configuration, we remove any
    default networks not specified (e.g. the default "bridge" network).
    """
    with network(subnet="10.247.197.96/27") as net:
        # Create a container with no specific network configuration. The only
        # networks connected will be the default ones.
        ret = docker_container.running(
            name=container_name,
            image=image,
            shutdown_timeout=1,
        )
        assert ret.result is True

        inspect_result = modules.docker.inspect_container(container_name)
        # Get the default network names
        default_networks = list(inspect_result["NetworkSettings"]["Networks"])

        # Re-run the state with an explicit network configuration. All of the
        # default networks should be disconnected.
        ret = docker_container.running(
            name=container_name,
            image=image,
            networks=[net.name],
            shutdown_timeout=1,
        )
        assert ret.result is True
        net_changes = ret.changes["container"]["Networks"]

        assert (
            f"Container '{container_name}' is already configured as specified."
            in ret.comment
        )

        updated_networks = modules.docker.inspect_container(container_name)[
            "NetworkSettings"
        ]["Networks"]

        for default_network in default_networks:
            assert f"Disconnected from network '{default_network}'." in ret.comment
            assert default_network in net_changes
            # We've tested that the state return is correct, but let's be extra
            # paranoid and check the actual connected networks.
            assert default_network not in updated_networks

        assert f"Connected to network '{net.name}'." in ret.comment


def test_run_with_onlyif(docker_container, container_name, image, modules):
    """
    Test docker_container.run with onlyif. The container should not run
    (and the state should return a True result) if the onlyif has a nonzero
    return code, but if the onlyif has a zero return code the container
    should run.
    If all items are True, the state should run
    """
    # Order should not matter
    test_cmds = (
        "/bin/false",  # single false
        ["/bin/true", "/bin/false"],  # false is last
        ["/bin/false", "/bin/true"],  # false is first
    )
    for cmd in test_cmds:
        log.debug("Trying %s", cmd)
        ret = docker_container.run(
            name=container_name,
            image=image,
            command="whoami",
            onlyif=cmd,
        )
        try:
            assert ret.result is True
            assert not ret.changes
            assert ret.comment.startswith("onlyif condition is false")
        except AssertionError:
            modules.docker.rm(container_name, force=True)
            raise

    for cmd in ("/bin/true", ["/bin/true", "ls /"]):
        log.debug("Trying %s", cmd)
        ret = docker_container.run(
            name=container_name,
            image=image,
            command="whoami",
            onlyif=cmd,
        )
        assert ret.result is True
        assert ret.changes["Logs"] == "root\n"
        assert ret.comment == "Container ran and exited with a return code of 0"
        modules.docker.rm(container_name, force=True)


def test_run_with_unless(docker_container, container_name, image, modules):
    """
    Test docker_container.run with unless. The container should not run
    (and the state should return a True result) if the unless has a zero
    return code, but if the unless has a nonzero return code the container
    should run.
    If any item is False, the state should run
    """
    # Test a single item and a list
    for cmd in ("/bin/true", ["/bin/true", "/bin/true"]):
        log.debug("Trying %s", cmd)
        ret = docker_container.run(
            name=container_name,
            image=image,
            command="whoami",
            unless=cmd,
        )
        try:
            assert ret.result is True
            assert not ret.changes
            assert ret.comment == "unless condition is true"
        except AssertionError:
            modules.docker.rm(container_name, force=True)
            raise

    # Order should not matter
    test_cmds = (
        "/bin/false",  # single false
        ["/bin/true", "/bin/false"],  # false is last
        ["/bin/false", "/bin/true"],  # false is first
    )
    for cmd in test_cmds:
        log.debug("Trying %s", cmd)
        ret = docker_container.run(
            name=container_name,
            image=image,
            command="whoami",
            unless=cmd,
        )
        assert ret.result is True
        assert ret.changes["Logs"] == "root\n"
        assert ret.comment == "Container ran and exited with a return code of 0"
        modules.docker.rm(container_name, force=True)


def test_run_with_creates(
    docker_container, container_name, image, tmp_path, subtests, modules
):
    """
    Test docker_container.run with creates. The container should not run
    (and the state should return a True result) if all of the files exist,
    but if if any of the files do not exist the container should run.
    """

    bad_file = str(tmp_path / "file-that-does-not-exist")
    good_file1 = tmp_path / "good1"
    good_file1.touch()
    good_file1 = str(good_file1)
    good_file2 = tmp_path / "good2"
    good_file2.touch()
    good_file2 = str(good_file2)

    with subtests.test(path=good_file1):
        try:
            log.debug("Trying %s", good_file1)
            ret = docker_container.run(
                name=container_name,
                image=image,
                command="whoami",
                creates=good_file1,
            )
            assert ret.result is True
            assert not ret.changes
            assert ret.comment == f"{good_file1} exists"
        finally:
            try:
                modules.docker.rm(container_name, force=True)
            except CommandExecutionError as exc:
                if "No such container" not in str(exc):
                    raise

    path = [good_file1, good_file2]
    with subtests.test(path=path):
        try:
            log.debug("Trying %s", path)
            ret = docker_container.run(
                name=container_name,
                image=image,
                command="whoami",
                creates=path,
            )
            assert ret.result is True
            assert not ret.changes
            assert ret.comment == "All files in creates exist"
        finally:
            try:
                modules.docker.rm(container_name, force=True)
            except CommandExecutionError as exc:
                if "No such container" not in str(exc):
                    raise

    for path in (bad_file, [good_file1, bad_file]):
        with subtests.test(path=path):
            try:
                log.debug("Trying %s", path)
                ret = docker_container.run(
                    name=container_name,
                    image=image,
                    command="whoami",
                    creates=path,
                )
                assert ret.result is True
                assert ret.changes["Logs"] == "root\n"
                assert ret.comment == "Container ran and exited with a return code of 0"
            finally:
                try:
                    modules.docker.rm(container_name, force=True)
                except CommandExecutionError as exc:
                    if "No such container" not in str(exc):
                        raise


def test_run_replace(docker_container, container_name, image):
    """
    Test the replace and force arguments to make sure they work properly
    """
    # Run once to create the container
    ret = docker_container.run(name=container_name, image=image, command="whoami")
    assert ret.result is True
    assert ret.changes["Logs"] == "root\n"
    assert ret.comment == "Container ran and exited with a return code of 0"

    # Run again with replace=False, this should fail
    ret = docker_container.run(
        name=container_name,
        image=image,
        command="whoami",
        replace=False,
    )
    assert ret.result is False
    assert not ret.changes
    assert ret.comment == (
        "Encountered error running container: Container '{}' exists. "
        "Run with replace=True to remove the existing container"
    ).format(container_name)

    # Run again with replace=True, this should proceed and there should be
    # a "Replaces" key in the changes dict to show that a container was
    # replaced.
    ret = docker_container.run(
        name=container_name,
        image=image,
        command="whoami",
        replace=True,
    )
    assert ret.result is True
    assert ret.changes["Logs"] == "root\n"
    assert "Replaces" in ret.changes
    assert ret.comment == "Container ran and exited with a return code of 0"


def test_run_force(docker_container, container_name, image):
    """
    Test the replace and force arguments to make sure they work properly
    """
    # Start up a container that will stay running
    ret = docker_container.running(name=container_name, image=image)
    assert ret.result is True

    # Run again with replace=True, this should fail because the container
    # is still running
    ret = docker_container.run(
        name=container_name,
        image=image,
        command="whoami",
        replace=True,
        force=False,
    )
    assert ret.result is False
    assert not ret.changes
    assert ret.comment == (
        "Encountered error running container: Container '{}' exists and is running. Run"
        " with replace=True and force=True to force removal of the existing container."
    ).format(container_name)

    # Run again with replace=True and force=True, this should proceed and
    # there should be a "Replaces" key in the changes dict to show that a
    # container was replaced.
    ret = docker_container.run(
        name=container_name,
        image=image,
        command="whoami",
        replace=True,
        force=True,
    )
    assert ret.result is True
    assert ret.changes["Logs"] == "root\n"
    assert "Replaces" in ret.changes
    assert ret.comment == "Container ran and exited with a return code of 0"


def test_run_failhard(docker_container, container_name, image, modules):
    """
    Test to make sure that we fail a state when the container exits with
    nonzero status if failhard is set to True, and that we don't when it is
    set to False.

    NOTE: We can't use RUNTIME_VARS.SHELL_FALSE_PATH here because the image
    we build on-the-fly here is based on busybox and does not include
    /usr/bin/false. Therefore, when the host machine running the tests
    has /usr/bin/false, it will not exist in the container and the Docker
    Engine API will cause an exception to be raised.
    """
    ret = docker_container.run(
        name=container_name,
        image=image,
        command="/bin/false",
        failhard=True,
    )
    assert ret.result is False
    assert ret.changes["Logs"] == ""
    assert ret.comment.startswith("Container ran and exited with a return code of")
    modules.docker.rm(container_name, force=True)

    ret = docker_container.run(
        name=container_name,
        image=image,
        command="/bin/false",
        failhard=False,
    )
    assert ret.result is True
    assert ret.changes["Logs"] == ""
    assert ret.comment.startswith("Container ran and exited with a return code of")
    modules.docker.rm(container_name, force=True)


def test_run_bg(docker_container, container_name, image, modules):
    """
    Test to make sure that if the container is run in the background, we do
    not include an ExitCode or Logs key in the return. Then check the logs
    for the container to ensure that it ran as expected.
    """
    ret = docker_container.run(
        name=container_name,
        image=image,
        command='sh -c "sleep 5 && whoami"',
        bg=True,
    )
    assert ret.result is True
    assert "Logs" not in ret.changes
    assert "ExitCode" not in ret.changes
    assert ret.comment == "Container was run in the background"

    # Now check the logs. The expectation is that the above asserts
    # completed during the 5-second sleep.
    assert modules.docker.logs(container_name, follow=True) == "root\n"



import logging
import os
import pathlib
import time

import attr
import pytest

import salt.channel.client
import salt.config
import salt.master
import salt.utils.files
import salt.utils.platform
import salt.utils.user

log = logging.getLogger(__name__)


@attr.s(slots=True, hash=True, frozen=True)
class UserInfo:
    master_config = attr.ib(repr=False, hash=False)
    username = attr.ib(init=False)
    key_file = attr.ib(init=False)
    key_path = attr.ib(init=False)
    key = attr.ib(init=False, repr=False)

    @username.default
    def _default_username(self):
        if not salt.utils.platform.is_windows():
            return self.master_config["user"]

        user = salt.utils.user.get_specific_user().replace("\\", "_")
        if user.startswith("sudo_"):
            user = user.split("sudo_")[-1]
        return user

    @key_file.default
    def _default_key_file(self):
        return f".{self.username}_key"

    @key_path.default
    def _default_key_path(self):
        return pathlib.Path(self.master_config["cachedir"]) / self.key_file

    @key.default
    def _default_key(self):
        with salt.utils.files.fopen(str(self.key_path)) as keyfd:
            return keyfd.read()


@pytest.fixture(scope="module")
def user_info(salt_master):
    return UserInfo(salt_master.config)


@pytest.fixture(scope="module")
def client_config(salt_minion, salt_master):
    opts = salt_minion.config.copy()
    opts.update(
        {
            "id": "root",
            "transport": salt_master.config["transport"],
            "auth_tries": 1,
            "auth_timeout": 5,
            "master_ip": "127.0.0.1",
            "master_port": salt_master.config["ret_port"],
            "master_uri": "tcp://127.0.0.1:{}".format(salt_master.config["ret_port"]),
        }
    )
    return opts


@pytest.fixture
def clear_channel(client_config):
    with salt.channel.client.ReqChannel.factory(
        client_config, crypt="clear"
    ) as channel:
        yield channel


def test_auth_info_not_allowed(clear_channel, user_info):
    assert hasattr(salt.master.ClearFuncs, "_prep_auth_info")

    msg = {"cmd": "_prep_auth_info"}
    rets = clear_channel.send(msg, timeout=15)
    ret_key = None
    for ret in rets:
        try:
            ret_key = ret[user_info.username]
            log.warning("User Key retrieved!!!:\n%s", ret)
            break
        except (TypeError, KeyError):
            pass
    assert ret_key != user_info.key, "Able to retrieve user key"


def test_pub_not_allowed(
    salt_master,
    clear_channel,
    tmp_path,
    event_listener,
    salt_minion,
    user_info,
    caplog,
):
    assert hasattr(salt.master.ClearFuncs, "_send_pub")
    tempfile = tmp_path / "evil_file"
    assert not tempfile.exists()
    jid = "202003100000000001"
    msg = {
        "cmd": "_send_pub",
        "fun": "file.write",
        "jid": jid,
        "arg": [str(tempfile), "evil contents"],
        "kwargs": {"show_jid": False, "show_timeout": False},
        "ret": "",
        "tgt": salt_minion.id,
        "tgt_type": "glob",
        "user": user_info.username,
    }
    timeout = 60
    start_time = time.time()
    expected_log_message = "Requested method not exposed: _send_pub"
    with caplog.at_level(logging.ERROR):
        clear_channel.send(msg, timeout=15)
        stop_time = start_time + timeout
        seen_records = []
        match_record = None
        while True:
            if match_record is not None:
                break
            if time.time() > stop_time:
                pytest.fail(
                    "Took more than {} seconds to confirm the presence of {!r} in the"
                    " logs".format(timeout, expected_log_message)
                )
            for record in caplog.records:
                if record in seen_records:
                    continue
                seen_records.append(record)
                if expected_log_message in str(record):
                    match_record = True
                    break
            time.sleep(0.5)

    # If we got the log message, we shouldn't get anything from the event bus
    expected_tag = f"salt/job/{jid}/*"
    event_pattern = (salt_master.id, expected_tag)
    events = event_listener.get_events([event_pattern], after_time=start_time)
    for event in events:
        pytest.fail(f"This event should't have gone through: {event}")

    assert not tempfile.exists(), "Evil file created"


def test_clearfuncs_config(salt_master, clear_channel, user_info):
    default_include_dir = pathlib.Path(salt_master.config["default_include"]).parent
    good_file_path = (
        pathlib.Path(salt_master.config_dir) / default_include_dir / "good.conf"
    )
    evil_file_path = pathlib.Path(salt_master.config_dir) / "evil.conf"
    assert not good_file_path.exists()
    assert not evil_file_path.exists()

    # assert good behavior
    good_msg = {
        "key": user_info.key,
        "cmd": "wheel",
        "fun": "config.update_config",
        "file_name": "good",
        "yaml_contents": "win: true",
    }
    ret = clear_channel.send(good_msg, timeout=5)
    assert "Wrote" in ret["data"]["return"]
    assert good_file_path.exists()
    good_file_path.unlink()
    try:
        evil_msg = {
            "key": user_info.key,
            "cmd": "wheel",
            "fun": "config.update_config",
            "file_name": "../evil",
            "yaml_contents": "win: true",
        }
        ret = clear_channel.send(evil_msg, timeout=5)
        assert not evil_file_path.exists(), "Wrote file via directory traversal"
        assert ret["data"]["return"] == "Invalid path"
    finally:
        if evil_file_path.exists():
            evil_file_path.unlink()


def test_fileroots_write(clear_channel, user_info, salt_master):
    state_tree_root_dir = pathlib.Path(salt_master.config["file_roots"]["base"][0])
    good_target = state_tree_root_dir / "good.txt"
    target_dir = state_tree_root_dir.parent
    bad_target = target_dir / "pwn.txt"

    # Good behaviour
    try:
        good_msg = {
            "key": user_info.key,
            "cmd": "wheel",
            "fun": "file_roots.write",
            "data": "win",
            "path": "good.txt",
            "saltenv": "base",
        }
        ret = clear_channel.send(good_msg, timeout=5)
        assert good_target.exists()
    finally:
        if good_target.exists():
            good_target.unlink()

    # Bad behaviour
    try:
        bad_msg = {
            "key": user_info.key,
            "cmd": "wheel",
            "fun": "file_roots.write",
            "data": "win",
            "path": os.path.join("..", "pwn.txt"),
            "saltenv": "base",
        }
        clear_channel.send(bad_msg, timeout=5)
        assert not bad_target.exists(), "Wrote file via directory traversal"
    finally:
        if bad_target.exists():
            bad_target.unlink()


def test_fileroots_read(clear_channel, user_info, salt_master):
    state_tree_root_dir = pathlib.Path(salt_master.config["file_roots"]["base"][0])
    # We can't use pathlib.Path.relative_to is does not behave the same as os.path.relpath
    # readpath = user_info.key_path.relative_to(state_tree_root_dir)
    readpath = os.path.relpath(str(user_info.key_path), str(state_tree_root_dir))
    relative_key_path = state_tree_root_dir / readpath
    log.debug("Master root_dir: %s", salt_master.config["root_dir"])
    log.debug("File Root: %s", state_tree_root_dir)
    log.debug("Key Path: %s", user_info.key_path)
    log.debug("Read Path: %s", readpath)
    log.debug("Relative Key Path: %s", relative_key_path)
    log.debug("Absolute Read Path: %s", relative_key_path.resolve())
    # If this asserion fails the test may need to be re-written
    assert relative_key_path.resolve() == user_info.key_path

    msg = {
        "key": user_info.key,
        "cmd": "wheel",
        "fun": "file_roots.read",
        "path": readpath,
        "saltenv": "base",
    }

    ret = clear_channel.send(msg, timeout=5)
    try:
        # When vulnerable this assertion will fail.
        assert (
            list(ret["data"]["return"][0].items())[0][1] != user_info.key
        ), "Read file via directory traversal"
    except IndexError:
        pass
    # If the vulnerability is fixed, no data will be returned.
    assert ret["data"]["return"] == []


def test_token(salt_master, salt_minion, clear_channel):
    tokensdir = pathlib.Path(salt_master.config["cachedir"]) / "tokens"
    assert tokensdir.is_dir()
    msg = {
        "arg": [],
        "cmd": "get_token",
        "token": str(pathlib.Path("..") / "minions" / salt_minion.id / "data.p"),
    }
    ret = clear_channel.send(msg, timeout=5)
    assert "pillar" not in ret, "Read minion data via directory traversal"



"""
Test the winrepo runner
"""

import textwrap

import pytest

import salt.runners.winrepo as winrepo
import salt.utils.files
import salt.utils.stringutils

# Can't use raw string with unicode_literals, since the \u in the uninstaller
# will be interpreted as a unicode code point and the interpreter will raise a
# SyntaxError.


@pytest.fixture
def winrepo_sls():
    _winrepo_sls = textwrap.dedent(
        """
    winscp_x86:
      5.7.5:
          full_name: 'WinSCP 5.7.5'
          installer: 'http://heanet.dl.sourceforge.net/project/winscp/WinSCP/5.7.5/winscp575setup.exe'
          install_flags: '/SP- /verysilent /norestart'
          uninstaller: '%PROGRAMFILES%\\WinSCP\\unins000.exe'
          uninstall_flags: '/verysilent'
          msiexec: False
          locale: en_US
          reboot: False
      5.7.4:
          full_name: 'WinSCP 5.7.4'
          installer: 'http://cznic.dl.sourceforge.net/project/winscp/WinSCP/5.7.4/winscp574setup.exe'
          install_flags: '/SP- /verysilent /norestart'
          uninstaller: '%PROGRAMFILES%\\WinSCP\\unins000.exe'
          uninstall_flags: '/verysilent'
          msiexec: False
          locale: en_US
          reboot: False
    """
    )
    return _winrepo_sls


@pytest.fixture
def winrepo_genrepo_data():
    _winrepo_genrepo_data = {
        "name_map": {"WinSCP 5.7.4": "winscp_x86", "WinSCP 5.7.5": "winscp_x86"},
        "repo": {
            "winscp_x86": {
                "5.7.5": {
                    "full_name": "WinSCP 5.7.5",
                    "installer": "http://heanet.dl.sourceforge.net/project/winscp/WinSCP/5.7.5/winscp575setup.exe",
                    "install_flags": "/SP- /verysilent /norestart",
                    "uninstaller": "%PROGRAMFILES%\\WinSCP\\unins000.exe",
                    "uninstall_flags": "/verysilent",
                    "msiexec": False,
                    "locale": "en_US",
                    "reboot": False,
                },
                "5.7.4": {
                    "full_name": "WinSCP 5.7.4",
                    "installer": "http://cznic.dl.sourceforge.net/project/winscp/WinSCP/5.7.4/winscp574setup.exe",
                    "install_flags": "/SP- /verysilent /norestart",
                    "uninstaller": "%PROGRAMFILES%\\WinSCP\\unins000.exe",
                    "uninstall_flags": "/verysilent",
                    "msiexec": False,
                    "locale": "en_US",
                    "reboot": False,
                },
            }
        },
    }
    return _winrepo_genrepo_data


@pytest.fixture
def winrepo_dir(tmp_path):
    dir_name = tmp_path / "winrepo_dir"
    dir_name.mkdir()
    return dir_name


@pytest.fixture
def extmods_dir(tmp_path):
    dir_name = tmp_path / "extmods_dir"
    dir_name.mkdir()
    return dir_name


@pytest.fixture
def winrepo_sls_dir(winrepo_dir):
    dir_name = winrepo_dir / "repo_sls"
    dir_name.mkdir()
    return dir_name


@pytest.fixture
def configure_loader_modules(winrepo_dir, extmods_dir):
    return {
        winrepo: {
            "__opts__": {
                "winrepo_cachefile": "winrepo.p",
                "optimization_order": [0, 1, 2],
                "renderer": "yaml",
                "renderer_blacklist": [],
                "renderer_whitelist": [],
                "file_roots": {"base": [str(winrepo_dir)]},
                "winrepo_dir": str(winrepo_dir),
                "extension_modules": str(extmods_dir),
            }
        }
    }


def test_genrepo(winrepo_sls, winrepo_genrepo_data, winrepo_sls_dir):
    """
    Test winrepo.genrepo runner
    """
    sls_file = str(winrepo_sls_dir / "wireshark.sls")
    # Add a winrepo SLS file
    with salt.utils.files.fopen(sls_file, "w") as fp_:
        fp_.write(salt.utils.stringutils.to_str(winrepo_sls))
    assert winrepo.genrepo() == winrepo_genrepo_data



"""
    :codeauthor: Gareth J. Greenaway <gareth@saltstack.com>
"""

import copy
import logging
import pathlib
import shutil
import tempfile
import textwrap

import pytest
from saltfactories.utils import random_string

import salt.config
import salt.ext.tornado
import salt.ext.tornado.testing
import salt.metaproxy.proxy
import salt.minion
import salt.syspaths
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class ProxyMinionTestCase(TestCase):
    @pytest.mark.slow_test
    def test_post_master_init_metaproxy_called(self):
        """
        Tests that when the _post_master_ini function is called, _metaproxy_call is also called.
        """

        mock_opts = salt.config.DEFAULT_MINION_OPTS.copy()
        mock_opts.update(salt.config.DEFAULT_PROXY_MINION_OPTS)
        mock_jid_queue = [123]
        proxy_minion = salt.minion.ProxyMinion(
            mock_opts,
            jid_queue=copy.copy(mock_jid_queue),
            io_loop=salt.ext.tornado.ioloop.IOLoop(),
        )
        mock_metaproxy_call = MagicMock()
        with patch(
            "salt.minion._metaproxy_call",
            return_value=mock_metaproxy_call,
            autospec=True,
        ):
            try:
                ret = proxy_minion._post_master_init("dummy_master")
                salt.minion._metaproxy_call.assert_called_once()
            finally:
                proxy_minion.destroy()

    @pytest.mark.slow_test
    def test_handle_decoded_payload_metaproxy_called(self):
        """
        Tests that when the _handle_decoded_payload function is called, _metaproxy_call is also called.
        """
        mock_opts = salt.config.DEFAULT_MINION_OPTS.copy()
        mock_opts.update(salt.config.DEFAULT_PROXY_MINION_OPTS)

        mock_data = {"fun": "foo.bar", "jid": 123}
        mock_jid_queue = [123]
        proxy_minion = salt.minion.ProxyMinion(
            mock_opts,
            jid_queue=copy.copy(mock_jid_queue),
            io_loop=salt.ext.tornado.ioloop.IOLoop(),
        )
        mock_metaproxy_call = MagicMock()
        with patch(
            "salt.minion._metaproxy_call",
            return_value=mock_metaproxy_call,
            autospec=True,
        ):
            try:
                ret = proxy_minion._handle_decoded_payload(mock_data).result()
                self.assertEqual(proxy_minion.jid_queue, mock_jid_queue)
                salt.minion._metaproxy_call.assert_called_once()
            finally:
                proxy_minion.destroy()

    @pytest.mark.slow_test
    def test_handle_payload_metaproxy_called(self):
        """
        Tests that when the _handle_payload function is called, _metaproxy_call is also called.
        """
        mock_opts = salt.config.DEFAULT_MINION_OPTS.copy()
        mock_opts.update(salt.config.DEFAULT_PROXY_MINION_OPTS)

        mock_data = {"fun": "foo.bar", "jid": 123}
        mock_jid_queue = [123]
        proxy_minion = salt.minion.ProxyMinion(
            mock_opts,
            jid_queue=copy.copy(mock_jid_queue),
            io_loop=salt.ext.tornado.ioloop.IOLoop(),
        )
        mock_metaproxy_call = MagicMock()
        with patch(
            "salt.minion._metaproxy_call",
            return_value=mock_metaproxy_call,
            autospec=True,
        ):
            try:
                ret = proxy_minion._handle_decoded_payload(mock_data).result()
                self.assertEqual(proxy_minion.jid_queue, mock_jid_queue)
                mock_metaproxy_call.assert_called_once()
            finally:
                proxy_minion.destroy()

    def test_proxy_config_default_include(self):
        """
        Tests that when the proxy_config function is called,
        for the proxy minion, eg. /etc/salt/proxy.d/<The-Proxy-ID>/*.conf
        """
        proxyid = random_string("proxy-")
        root_dir = pathlib.Path(tempfile.mkdtemp(dir=RUNTIME_VARS.TMP))
        self.addCleanup(shutil.rmtree, str(root_dir), ignore_errors=True)
        conf_dir = root_dir / "conf"
        conf_file = conf_dir / "proxy"
        conf_d_dir = conf_dir / "proxy.d"
        proxy_conf_d = conf_d_dir / proxyid
        proxy_conf_d.mkdir(parents=True)

        with salt.utils.files.fopen(str(conf_file), "w") as wfh:
            wfh.write(
                textwrap.dedent(
                    """\
                    id: {id}
                    root_dir: {root_dir}
                    pidfile: run/proxy.pid
                    pki_dir: pki
                    cachedir: cache
                    sock_dir: run/proxy
                    log_file: logs/proxy.log
                    """.format(
                        id=proxyid, root_dir=root_dir
                    )
                )
            )

        with salt.utils.files.fopen(str(proxy_conf_d / "_schedule.conf"), "w") as wfh:
            wfh.write(
                textwrap.dedent(
                    """\
                    schedule:
                      test_job:
                        args: [arg1, arg2]
                        enabled: true
                        function: test.arg
                        jid_include: true
                        kwargs: {key1: value1, key2: value2}
                        maxrunning: 1
                        name: test_job
                        return_job: false
                    """
                )
            )
        opts = salt.config.proxy_config(
            str(conf_file),
            minion_id=proxyid,
            cache_minion_id=False,
        )
        self.assertIn("schedule", opts)
        self.assertIn("test_job", opts["schedule"])



"""
unittests for terraform roster
"""

import pathlib

import pytest

from salt.roster import terraform
from salt.utils import roster_matcher
from tests.support.mock import patch


@pytest.fixture
def roster_file():
    return pathlib.Path(__file__).parent / "terraform.data" / "terraform.tfstate"


@pytest.fixture
def roster_file_new():
    return pathlib.Path(__file__).parent / "terraform.data" / "terraform-new.tfstate"


@pytest.fixture
def pki_dir():
    return pathlib.Path(__file__).parent / "terraform.data"


@pytest.fixture
def configure_loader_modules(roster_file, pki_dir):
    return {
        terraform: {
            "__utils__": {
                "roster_matcher.targets": roster_matcher.targets,
            },
            "__opts__": {
                "roster_file": str(roster_file),
                "pki_dir": str(pki_dir),
            },
        },
        roster_matcher: {},
    }


def test_default_output(pki_dir):
    """
    Test the output of a fixture tfstate file which contains libvirt
    resources.
    """
    expected_result = {
        "db0": {
            "host": "192.168.122.174",
            "user": "root",
            "passwd": "dbpw",
            "tty": True,
            "priv": str(pki_dir / "ssh" / "salt-ssh.rsa"),
        },
        "db1": {
            "host": "192.168.122.190",
            "user": "root",
            "passwd": "dbpw",
            "tty": True,
            "priv": str(pki_dir / "ssh" / "salt-ssh.rsa"),
        },
        "web0": {
            "host": "192.168.122.106",
            "user": "root",
            "passwd": "linux",
            "timeout": 22,
            "priv": str(pki_dir / "ssh" / "salt-ssh.rsa"),
        },
        "web1": {
            "host": "192.168.122.235",
            "user": "root",
            "passwd": "linux",
            "timeout": 22,
            "priv": str(pki_dir / "ssh" / "salt-ssh.rsa"),
        },
    }

    ret = terraform.targets("*")
    assert expected_result == ret


def test_default_matching(pki_dir):
    """
    Test the output of a fixture tfstate file which contains libvirt
    resources using matching
    """
    expected_result = {
        "web0": {
            "host": "192.168.122.106",
            "user": "root",
            "passwd": "linux",
            "timeout": 22,
            "priv": str(pki_dir / "ssh" / "salt-ssh.rsa"),
        },
        "web1": {
            "host": "192.168.122.235",
            "user": "root",
            "passwd": "linux",
            "timeout": 22,
            "priv": str(pki_dir / "ssh" / "salt-ssh.rsa"),
        },
    }

    ret = terraform.targets("*web*")
    assert expected_result == ret


def test_defaults_new(pki_dir, roster_file_new):
    """
    Test the output of a fixture tfstate file which contains libvirt
    resources using matching
    """
    expected_result = {
        "web0": {
            "host": "192.168.122.106",
            "user": "root",
            "passwd": "linux",
            "timeout": 22,
            "priv": str(pki_dir / "ssh" / "salt-ssh.rsa"),
        },
        "web1": {
            "host": "192.168.122.107",
            "user": "root",
            "passwd": "linux",
            "timeout": 22,
            "priv": str(pki_dir / "ssh" / "salt-ssh.rsa"),
        },
    }

    with patch.dict(terraform.__opts__, {"roster_file": str(roster_file_new)}):
        ret = terraform.targets("*")
        assert expected_result == ret


def test_defaults_new_matching(pki_dir, roster_file_new):
    """
    Test the output of a fixture tfstate file which contains libvirt
    resources using matching
    """
    expected_result = {
        "web1": {
            "host": "192.168.122.107",
            "user": "root",
            "passwd": "linux",
            "timeout": 22,
            "priv": str(pki_dir / "ssh" / "salt-ssh.rsa"),
        },
    }

    with patch.dict(terraform.__opts__, {"roster_file": str(roster_file_new)}):
        ret = terraform.targets("*1")
        assert expected_result == ret


def test_correct_handler_called_old():
    old_mock = patch.object(terraform, "_do__parse_old_state_file")
    new_mock = patch.object(terraform, "_do_parse_new_state_file")
    with old_mock as _old, new_mock as _new:
        terraform.targets("*")
        _old.assert_called_once()
        _new.assert_not_called()


def test_correct_handler_called_new(roster_file_new):
    old_mock = patch.object(terraform, "_do__parse_old_state_file")
    new_mock = patch.object(terraform, "_do_parse_new_state_file")
    with old_mock as _old, new_mock as _new:
        with patch.dict(terraform.__opts__, {"roster_file": str(roster_file_new)}):
            terraform.targets("*")
            _old.assert_not_called()
            _new.assert_called_once()



"""
The match module allows for match routines to be run and determine target specs
"""

import copy
import inspect
import logging
import sys
from collections.abc import Mapping

import salt.loader
from salt.defaults import DEFAULT_TARGET_DELIM
from salt.exceptions import SaltException
from salt.utils.decorators.jinja import jinja_global

__func_alias__ = {"list_": "list"}

log = logging.getLogger(__name__)


def _load_matchers():
    """
    Store matchers in __context__ so they're only loaded once
    """
    __context__["matchers"] = salt.loader.matchers(__opts__)


def compound(tgt, minion_id=None):
    """
    Return True if the minion ID matches the given compound target

    minion_id
        Specify the minion ID to match against the target expression

        .. versionadded:: 2014.7.0

    CLI Example:

    .. code-block:: bash

        salt '*' match.compound 'L@cheese,foo and *'
    """
    if minion_id is not None:
        if not isinstance(minion_id, str):
            minion_id = str(minion_id)
    if "matchers" not in __context__:
        _load_matchers()
    try:
        ret = __context__["matchers"]["compound_match.match"](
            tgt, opts=__opts__, minion_id=minion_id
        )
    except Exception as exc:  # pylint: disable=broad-except
        log.exception(exc)
        ret = False

    return ret


def ipcidr(tgt):
    """
    Return True if the minion matches the given ipcidr target

    CLI Example:

    .. code-block:: bash

        salt '*' match.ipcidr '192.168.44.0/24'

    delimiter
    Pillar Example:

    .. code-block:: yaml

       '172.16.0.0/12':
         - match: ipcidr
         - nodeclass: internal

    """
    if "matchers" not in __context__:
        _load_matchers()
    try:
        return __context__["matchers"]["ipcidr_match.match"](tgt, opts=__opts__)
    except Exception as exc:  # pylint: disable=broad-except
        log.exception(exc)
        return False


def pillar_pcre(tgt, delimiter=DEFAULT_TARGET_DELIM):
    """
    Return True if the minion matches the given pillar_pcre target. The
    ``delimiter`` argument can be used to specify a different delimiter.

    CLI Example:

    .. code-block:: bash

        salt '*' match.pillar_pcre 'cheese:(swiss|american)'
        salt '*' match.pillar_pcre 'clone_url|https://github\\.com/.*\\.git' delimiter='|'

    delimiter
        Specify an alternate delimiter to use when traversing a nested dict

        .. versionadded:: 2014.7.0

    delim
        Specify an alternate delimiter to use when traversing a nested dict

        .. versionadded:: 0.16.4
        .. deprecated:: 2015.8.0
    """
    if "matchers" not in __context__:
        _load_matchers()
    try:
        return __context__["matchers"]["pillar_pcre_match.match"](
            tgt, delimiter=delimiter, opts=__opts__
        )
    except Exception as exc:  # pylint: disable=broad-except
        log.exception(exc)
        return False


def pillar(tgt, delimiter=DEFAULT_TARGET_DELIM):
    """
    Return True if the minion matches the given pillar target. The
    ``delimiter`` argument can be used to specify a different delimiter.

    CLI Example:

    .. code-block:: bash

        salt '*' match.pillar 'cheese:foo'
        salt '*' match.pillar 'clone_url|https://github.com/saltstack/salt.git' delimiter='|'

    delimiter
        Specify an alternate delimiter to use when traversing a nested dict

        .. versionadded:: 2014.7.0

    delim
        Specify an alternate delimiter to use when traversing a nested dict

        .. versionadded:: 0.16.4
        .. deprecated:: 2015.8.0
    """
    if "matchers" not in __context__:
        _load_matchers()
    try:
        return __context__["matchers"]["pillar_match.match"](
            tgt, delimiter=delimiter, opts=__opts__
        )
    except Exception as exc:  # pylint: disable=broad-except
        log.exception(exc)
        return False


def data(tgt):
    """
    Return True if the minion matches the given data target

    CLI Example:

    .. code-block:: bash

        salt '*' match.data 'spam:eggs'
    """
    if "matchers" not in __context__:
        _load_matchers()
    try:
        return __context__["matchers"]["data_match.match"](tgt, opts=__opts__)
    except Exception as exc:  # pylint: disable=broad-except
        log.exception(exc)
        return False


def grain_pcre(tgt, delimiter=DEFAULT_TARGET_DELIM):
    """
    Return True if the minion matches the given grain_pcre target. The
    ``delimiter`` argument can be used to specify a different delimiter.

    CLI Example:

    .. code-block:: bash

        salt '*' match.grain_pcre 'os:Fedo.*'
        salt '*' match.grain_pcre 'ipv6|2001:.*' delimiter='|'

    delimiter
        Specify an alternate delimiter to use when traversing a nested dict

        .. versionadded:: 2014.7.0

    delim
        Specify an alternate delimiter to use when traversing a nested dict

        .. versionadded:: 0.16.4
        .. deprecated:: 2015.8.0
    """
    if "matchers" not in __context__:
        _load_matchers()
    try:
        return __context__["matchers"]["grain_pcre_match.match"](
            tgt, delimiter=delimiter, opts=__opts__
        )
    except Exception as exc:  # pylint: disable=broad-except
        log.exception(exc)
        return False


def grain(tgt, delimiter=DEFAULT_TARGET_DELIM):
    """
    Return True if the minion matches the given grain target. The ``delimiter``
    argument can be used to specify a different delimiter.

    CLI Example:

    .. code-block:: bash

        salt '*' match.grain 'os:Ubuntu'
        salt '*' match.grain 'ipv6|2001:db8::ff00:42:8329' delimiter='|'

    delimiter
        Specify an alternate delimiter to use when traversing a nested dict

        .. versionadded:: 2014.7.0

    delim
        Specify an alternate delimiter to use when traversing a nested dict

        .. versionadded:: 0.16.4
        .. deprecated:: 2015.8.0
    """
    if "matchers" not in __context__:
        _load_matchers()
    try:
        return __context__["matchers"]["grain_match.match"](
            tgt, delimiter=delimiter, opts=__opts__
        )
    except Exception as exc:  # pylint: disable=broad-except
        log.exception(exc)
        return False


def list_(tgt, minion_id=None):
    """
    Return True if the minion ID matches the given list target

    minion_id
        Specify the minion ID to match against the target expression

        .. versionadded:: 2014.7.0

    CLI Example:

    .. code-block:: bash

        salt '*' match.list 'server1,server2'
    """
    if minion_id is not None:
        if not isinstance(minion_id, str):
            minion_id = str(minion_id)
    if "matchers" not in __context__:
        _load_matchers()
    try:
        return __context__["matchers"]["list_match.match"](
            tgt, opts=__opts__, minion_id=minion_id
        )
    except Exception as exc:  # pylint: disable=broad-except
        log.exception(exc)
        return False


def pcre(tgt, minion_id=None):
    """
    Return True if the minion ID matches the given pcre target

    minion_id
        Specify the minion ID to match against the target expression

        .. versionadded:: 2014.7.0

    CLI Example:

    .. code-block:: bash

        salt '*' match.pcre '.*'
    """
    if minion_id is not None:
        if not isinstance(minion_id, str):
            minion_id = str(minion_id)
    if "matchers" not in __context__:
        _load_matchers()
    try:
        return __context__["matchers"]["pcre_match.match"](
            tgt, opts=__opts__, minion_id=minion_id
        )
    except Exception as exc:  # pylint: disable=broad-except
        log.exception(exc)
        return False


def glob(tgt, minion_id=None):
    """
    Return True if the minion ID matches the given glob target

    minion_id
        Specify the minion ID to match against the target expression

        .. versionadded:: 2014.7.0

    CLI Example:

    .. code-block:: bash

        salt '*' match.glob '*'
    """
    if minion_id is not None:
        if not isinstance(minion_id, str):
            minion_id = str(minion_id)
    if "matchers" not in __context__:
        _load_matchers()

    try:
        return __context__["matchers"]["glob_match.match"](
            tgt, opts=__opts__, minion_id=minion_id
        )
    except Exception as exc:  # pylint: disable=broad-except
        log.exception(exc)
        return False


def filter_by(
    lookup,
    tgt_type="compound",
    minion_id=None,
    merge=None,
    merge_lists=False,
    default="default",
):
    """
    Return the first match in a dictionary of target patterns

    .. versionadded:: 2014.7.0

    CLI Example:

    .. code-block:: bash

        salt '*' match.filter_by '{foo*: Foo!, bar*: Bar!}' minion_id=bar03

    Pillar Example:

    .. code-block:: jinja

        # Filter the data for the current minion into a variable:
        {% set roles = salt['match.filter_by']({
            'web*': ['app', 'caching'],
            'db*': ['db'],
        }, minion_id=grains['id'], default='web*') %}

        # Make the filtered data available to Pillar:
        roles: {{ roles | yaml() }}
    """
    expr_funcs = dict(
        inspect.getmembers(sys.modules[__name__], predicate=inspect.isfunction)
    )

    for key in lookup:
        params = (key, minion_id) if minion_id else (key,)
        if expr_funcs[tgt_type](*params):
            if merge:
                if not isinstance(merge, Mapping):
                    raise SaltException(
                        "filter_by merge argument must be a dictionary."
                    )

                if lookup[key] is None:
                    return merge
                else:
                    salt.utils.dictupdate.update(
                        lookup[key], copy.deepcopy(merge), merge_lists=merge_lists
                    )

            return lookup[key]

    return lookup.get(default, None)


def search_by(lookup, tgt_type="compound", minion_id=None):
    """
    Search a dictionary of target strings for matching targets

    This is the inverse of :py:func:`match.filter_by
    <salt.modules.match.filter_by>` and allows matching values instead of
    matching keys. A minion can be matched by multiple entries.

    .. versionadded:: 2017.7.0

    CLI Example:

    .. code-block:: bash

        salt '*' match.search_by '{web: [node1, node2], db: [node2, node]}'

    Pillar Example:

    .. code-block:: jinja

        {% set roles = salt.match.search_by({
            'web': ['G@os_family:Debian not nodeX'],
            'db': ['L@node2,node3 and G@datacenter:west'],
            'caching': ['node3', 'node4'],
        }) %}

        # Make the filtered data available to Pillar:
        roles: {{ roles | yaml() }}
    """
    expr_funcs = dict(
        inspect.getmembers(sys.modules[__name__], predicate=inspect.isfunction)
    )

    matches = []
    for key, target_list in lookup.items():
        for target in target_list:
            params = (target, minion_id) if minion_id else (target,)
            if expr_funcs[tgt_type](*params):
                matches.append(key)

    return matches or None


@jinja_global("ifelse")
def ifelse(
    *args,
    tgt_type="compound",
    minion_id=None,
    merge=None,
    merge_lists=False,
):
    """
    .. versionadded:: 3006.0

    Evaluate each pair of arguments up to the last one as a (matcher, value)
    tuple, returning ``value`` if matched.  If none match, returns the last
    argument.

    The ``ifelse`` function is like a multi-level if-else statement. It was
    inspired by CFEngine's ``ifelse`` function which in turn was inspired by
    Oracle's ``DECODE`` function. It must have an odd number of arguments (from
    1 to N). The last argument is the default value, like the ``else`` clause in
    standard programming languages. Every pair of arguments before the last one
    are evaluated as a pair. If the first one evaluates true then the second one
    is returned, as if you had used the first one in a compound match
    expression. Boolean values can also be used as the first item in a pair,
    as it will be translated to a match that will always match ("*") or never
    match ("SALT_IFELSE_MATCH_NOTHING") a target system.

    This is essentially another way to express the ``filter_by`` functionality
    in way that's familiar to CFEngine or Oracle users. Consider using
    ``filter_by`` unless this function fits your workflow.

    CLI Example:

    .. code-block:: bash

        salt '*' match.ifelse 'foo*' 'Foo!' 'bar*' 'Bar!' minion_id=bar03
    """
    if len(args) % 2 == 0:
        raise SaltException("The ifelse function must have an odd number of arguments!")
    elif len(args) == 1:
        return args[0]

    default_key = "SALT_IFELSE_FUNCTION_DEFAULT"

    keys = list(args[::2])
    for idx, key in enumerate(keys):
        if key is True:
            keys[idx] = "*"
        elif key is False:
            keys[idx] = "SALT_IFELSE_MATCH_NOTHING"

    lookup = dict(zip(keys, args[1::2]))
    lookup.update({default_key: args[-1]})

    return filter_by(
        lookup=lookup,
        tgt_type=tgt_type,
        minion_id=minion_id,
        merge=merge,
        merge_lists=merge_lists,
        default=default_key,
    )



"""
Interface with Helm

:depends: pyhelm_ Python package

.. _pyhelm: https://pypi.org/project/pyhelm/

.. note::
    This module use the helm-cli. The helm-cli binary have to be present in your Salt-Minion path.

Helm-CLI vs Salt-Modules
------------------------

This module is a wrapper of the helm binary.
All helm v3.0 command are implemented.

To install a chart with the helm-cli:

.. code-block:: bash

    helm install grafana stable/grafana --wait --values /path/to/values.yaml


To install a chart with the Salt-Module:

.. code-block:: bash

    salt '*' helm.install grafana stable/grafana values='/path/to/values.yaml' flags="['wait']"


Detailed Function Documentation
-------------------------------
"""

import copy
import logging
import re

from salt.exceptions import CommandExecutionError
from salt.serializers import json

log = logging.getLogger(__name__)

# Don't shadow built-in's.
__func_alias__ = {
    "help_": "help",
    "list_": "list",
}


def _prepare_cmd(binary="helm", commands=None, flags=None, kvflags=None):
    """

    :param binary:
    :param commands:
    :param flags:
    :param kvflags:
    :return:
    """
    if commands is None:
        commands = []
    if flags is None:
        flags = []
    else:
        flags = copy.deepcopy(flags)
    if kvflags is None:
        kvflags = {}
    else:
        kvflags = copy.deepcopy(kvflags)

    cmd = (binary,)

    for command in commands:
        cmd += (command,)

    for arg in flags:
        if not re.search(r"^--.*", arg):
            arg = "--" + arg
        cmd += (arg,)

    for key, val in kvflags.items():
        if not re.search(r"^--.*", key):
            key = "--" + key
        if key == "--set" and isinstance(val, list):
            for set_val in val:
                cmd += (
                    key,
                    set_val,
                )
        else:
            cmd += (
                key,
                val,
            )

    return cmd


def _exec_cmd(commands=None, flags=None, kvflags=None):
    """

    :param commands:
    :param flags:
    :param kvflags:
    :return:
    """
    cmd = _prepare_cmd(commands=commands, flags=flags, kvflags=kvflags)
    cmd_string = " ".join(cmd)

    try:
        result = __salt__["cmd.run_all"](cmd=cmd)
        result.update({"cmd": cmd_string})
    except CommandExecutionError as err:
        result = {"retcode": -1, "stdout": "", "stderr": err, "cmd": cmd_string}
        log.error(result)

    return result


def _exec_true_return(commands=None, flags=None, kvflags=None):
    """

    :param commands:
    :param flags:
    :param kvflags:
    :return:
    """
    cmd_result = _exec_cmd(commands=commands, flags=flags, kvflags=kvflags)
    if cmd_result.get("retcode", -1) == 0:
        result = True
    else:
        result = cmd_result.get("stderr", "")
    return result


def _exec_string_return(commands=None, flags=None, kvflags=None):
    """

    :param commands:
    :param flags:
    :param kvflags:
    :return:
    """
    cmd_result = _exec_cmd(commands=commands, flags=flags, kvflags=kvflags)
    if cmd_result.get("retcode", -1) == 0:
        result = cmd_result.get("stdout", "")
    else:
        result = cmd_result.get("stderr", "")
    return result


def _exec_dict_return(commands=None, flags=None, kvflags=None):
    """

    :param commands:
    :param flags:
    :param kvflags:
    :return:
    """
    if kvflags is None:
        kvflags = {}
    if not ("output" in kvflags.keys() or "--output" in kvflags.keys()):
        kvflags.update({"output": "json"})
    cmd_result = _exec_cmd(commands=commands, flags=flags, kvflags=kvflags)
    if cmd_result.get("retcode", -1) == 0:
        if kvflags.get("output") == "json" or kvflags.get("--output") == "json":
            result = json.deserialize(cmd_result.get("stdout", ""))
        else:
            result = cmd_result.get("stdout", "")
    else:
        result = cmd_result.get("stderr", "")
    return result


def completion(shell, flags=None, kvflags=None):
    """
    Generate auto-completions script for Helm for the specified shell (bash or zsh).
    Return the shell auto-completion content.

    shell
        (string) One of ['bash', 'zsh'].

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.completion bash

    """
    return _exec_string_return(
        commands=["completion", shell], flags=flags, kvflags=kvflags
    )


def create(name, flags=None, kvflags=None):
    """
    Creates a chart directory along with the common files and directories used in a chart.
    Return True if succeed, else the error message.

    name
        (string) The chart name to create.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.create NAME

    """
    return _exec_true_return(commands=["create", name], flags=flags, kvflags=kvflags)


def dependency_build(chart, flags=None, kvflags=None):
    """
    Build out the charts/ directory from the Chart.lock file.
    Return True if succeed, else the error message.

    chart
        (string) The chart name to build dependency.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.dependency_build CHART

    """
    return _exec_true_return(
        commands=["dependency", "build", chart], flags=flags, kvflags=kvflags
    )


def dependency_list(chart, flags=None, kvflags=None):
    """
    List all of the dependencies declared in a chart.
    Return chart dependencies if succeed, else the error message.

    chart
        (string) The chart name to list dependency.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.dependency_list CHART

    """
    return _exec_string_return(
        commands=["dependency", "list", chart], flags=flags, kvflags=kvflags
    )


def dependency_update(chart, flags=None, kvflags=None):
    """
    Update the on-disk dependencies to mirror Chart.yaml.
    Return True if succeed, else the error message.

    chart
        (string) The chart name to update dependency.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.dependency_update CHART

    """
    return _exec_true_return(
        commands=["dependency", "update", chart], flags=flags, kvflags=kvflags
    )


def env(flags=None, kvflags=None):
    """
    Prints out all the environment information in use by Helm.
    Return Helm environments variables if succeed, else the error message.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.env

    """
    return _exec_string_return(commands=["env"], flags=flags, kvflags=kvflags)


def get_all(release, flags=None, kvflags=None):
    """
    Prints a human readable collection of information about the notes, hooks, supplied values, and generated manifest file of the given release.
    Return release information if succeed, else the error message.

    release
        (string) Release name to get information from.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.get_all RELEASE

    """
    return _exec_string_return(
        commands=["get", "all", release], flags=flags, kvflags=kvflags
    )


def get_hooks(release, flags=None, kvflags=None):
    """
    Prints a human readable collection of information about the hooks of the given release.
    Return release hooks information if succeed, else the error message.

    release
        (string) Release name to get hooks information from.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.get_hooks RELEASE

    """
    return _exec_string_return(
        commands=["get", "hooks", release], flags=flags, kvflags=kvflags
    )


def get_manifest(release, flags=None, kvflags=None):
    """
    Prints a human readable collection of information about the manifest of the given release.
    Return release manifest information if succeed, else the error message.

    release
        (string) Release name to get manifest information from.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.get_manifest RELEASE

    """
    return _exec_string_return(
        commands=["get", "manifest", release], flags=flags, kvflags=kvflags
    )


def get_notes(release, flags=None, kvflags=None):
    """
    Prints a human readable collection of information about the notes of the given release.
    Return release notes information if succeed, else the error message.

    release
        (string) Release name to get notes information from.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.get_notes RELEASE

    """
    return _exec_string_return(
        commands=["get", "notes", release], flags=flags, kvflags=kvflags
    )


def get_values(release, flags=None, kvflags=None):
    """
    Prints a human readable collection of information about the values of the given release.
    Return release values information if succeed, else the error message.

    release
        (string) Release name to get values information from.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.get_values RELEASE

        # In YAML format
        salt '*' helm.get_values RELEASE kvflags="{'output': 'yaml'}"

    """
    return _exec_dict_return(
        commands=["get", "values", release], flags=flags, kvflags=kvflags
    )


def help_(command, flags=None, kvflags=None):
    """
    Provides help for any command in the application.
    Return the full help if succeed, else the error message.

    command
        (string) Command to get help. ex: 'get'

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.help COMMAND

    """
    return _exec_string_return(commands=["help", command], flags=flags, kvflags=kvflags)


def history(release, flags=None, kvflags=None):
    """
    Prints historical revisions for a given release.
    Return release historic if succeed, else the error message.

    release
        (string) Release name to get history from.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.history RELEASE

        # In YAML format
        salt '*' helm.history RELEASE kvflags="{'output': 'yaml'}"

    """
    return _exec_dict_return(
        commands=["history", release], flags=flags, kvflags=kvflags
    )


def install(
    release,
    chart,
    values=None,
    version=None,
    namespace=None,
    set=None,
    flags=None,
    kvflags=None,
):
    """
    Installs a chart archive.
    Return True if succeed, else the error message.

    release
        (string) Release name to get values information from.

    chart
        (string) Chart name to install.

    values
        (string) Absolute path to the values.yaml file.

    version
        (string) The exact chart version to install. If this is not specified, the latest version is installed.

    namespace
        (string) The namespace scope for this request.

    set
        (string or list) Set a values on the command line.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.install RELEASE CHART

        # With values file.
        salt '*' helm.install RELEASE CHART values='/path/to/values.yaml'

    """
    if values:
        if kvflags:
            kvflags.update({"values": values})
        else:
            kvflags = {"values": values}
    if version:
        if kvflags:
            kvflags.update({"version": version})
        else:
            kvflags = {"version": version}
    if namespace:
        if kvflags:
            kvflags.update({"namespace": namespace})
        else:
            kvflags = {"namespace": namespace}
    if set:
        if kvflags:
            kvflags.update({"set": set})
        else:
            kvflags = {"set": set}
    return _exec_true_return(
        commands=["install", release, chart], flags=flags, kvflags=kvflags
    )


def lint(path, values=None, namespace=None, set=None, flags=None, kvflags=None):
    """
    Takes a path to a chart and runs a series of tests to verify that the chart is well-formed.
    Return True if succeed, else the error message.

    path
        (string) The path to the chart to lint.

    values
        (string) Absolute path to the values.yaml file.

    namespace
        (string) The namespace scope for this request.

    set
        (string or list) Set a values on the command line.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.lint PATH

    """
    if values:
        if kvflags:
            kvflags.update({"values": values})
        else:
            kvflags = {"values": values}
    if namespace:
        if kvflags:
            kvflags.update({"namespace": namespace})
        else:
            kvflags = {"namespace": namespace}
    if set:
        if kvflags:
            kvflags.update({"set": set})
        else:
            kvflags = {"set": set}
    return _exec_true_return(commands=["lint", path], flags=flags, kvflags=kvflags)


def list_(namespace=None, flags=None, kvflags=None):
    """
    Lists all of the releases. By default, it lists only releases that are deployed or failed.
    Return the list of release if succeed, else the error message.

    namespace
        (string) The namespace scope for this request.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.list

        # In YAML format
        salt '*' helm.list kvflags="{'output': 'yaml'}"

    """
    if namespace:
        if kvflags:
            kvflags.update({"namespace": namespace})
        else:
            kvflags = {"namespace": namespace}
    return _exec_dict_return(commands=["list"], flags=flags, kvflags=kvflags)


def package(chart, flags=None, kvflags=None):
    """
    Packages a chart into a versioned chart archive file. If a path is given, this will look at that path for a chart
    (which must contain a Chart.yaml file) and then package that directory.
    Return True if succeed, else the error message.

    chart
        (string) Chart name to package. Can be an absolute path.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.package CHART

        # With destination path.
        salt '*' helm.package CHART kvflags="{'destination': '/path/to/the/package'}"

    """
    return _exec_true_return(commands=["package", chart], flags=flags, kvflags=kvflags)


def plugin_install(path, flags=None, kvflags=None):
    """
    Install a Helm plugin from a url to a VCS repo or a local path.
    Return True if succeed, else the error message.

    path
        (string) Path to the local plugin. Can be an url.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.plugin_install PATH

    """
    return _exec_true_return(
        commands=["plugin", "install", path], flags=flags, kvflags=kvflags
    )


def plugin_list(flags=None, kvflags=None):
    """
    List installed Helm plugins.
    Return the plugin list if succeed, else the error message.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.plugin_list

    """
    return _exec_string_return(
        commands=["plugin", "list"], flags=flags, kvflags=kvflags
    )


def plugin_uninstall(plugin, flags=None, kvflags=None):
    """
    Uninstall a Helm plugin.
    Return True if succeed, else the error message.

    plugin
        (string) The plugin to uninstall.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.plugin_uninstall PLUGIN

    """
    return _exec_true_return(
        commands=["plugin", "uninstall", plugin], flags=flags, kvflags=kvflags
    )


def plugin_update(plugin, flags=None, kvflags=None):
    """
    Update a Helm plugin.
    Return True if succeed, else the error message.

    plugin
        (string) The plugin to update.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.plugin_update PLUGIN

    """
    return _exec_true_return(
        commands=["plugin", "update", plugin], flags=flags, kvflags=kvflags
    )


def pull(pkg, flags=None, kvflags=None):
    """
    Retrieve a package from a package repository, and download it locally.
    Return True if succeed, else the error message.

    pkg
        (string) The package to pull. Can be url or repo/chartname.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.pull PKG

        # With destination path to write the chart.
        salt '*' helm.pull PKG kvflags="{'destination': '/path/to/the/chart'}"

    """
    return _exec_true_return(commands=["pull", pkg], flags=flags, kvflags=kvflags)


def repo_add(name, url, namespace=None, flags=None, kvflags=None):
    """
    Add a chart repository.
    Return True if succeed, else the error message.

    name
        (string) The local name of the repository to install. Have to be unique.

    url
        (string) The url to the repository.

    namespace
        (string) The namespace scope for this request.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.repo_add NAME URL

    """
    if namespace:
        if kvflags:
            kvflags.update({"namespace": namespace})
        else:
            kvflags = {"namespace": namespace}
    return _exec_true_return(
        commands=["repo", "add", name, url], flags=flags, kvflags=kvflags
    )


def repo_index(directory, namespace=None, flags=None, kvflags=None):
    """
    Read the current directory and generate an index file based on the charts found.
    Return True if succeed, else the error message.

    directory
        (string) The path to the index.

    namespace
        (string) The namespace scope for this request.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.index DIRECTORY

    """
    if namespace:
        if kvflags:
            kvflags.update({"namespace": namespace})
        else:
            kvflags = {"namespace": namespace}
    return _exec_true_return(
        commands=["repo", "index", directory], flags=flags, kvflags=kvflags
    )


def repo_list(namespace=None, flags=None, kvflags=None):
    """
    List a chart repository.
    Return the repository list if succeed, else the error message.

    namespace
        (string) The namespace scope for this request.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.repo_list

        # In YAML format
        salt '*' helm.repo_list kvflags="{'output': 'yaml'}"

    """
    if namespace:
        if kvflags:
            kvflags.update({"namespace": namespace})
        else:
            kvflags = {"namespace": namespace}
    return _exec_dict_return(commands=["repo", "list"], flags=flags, kvflags=kvflags)


def repo_remove(name, namespace=None, flags=None, kvflags=None):
    """
    Remove a chart repository.
    Return True if succeed, else the error message.

    name
        (string) The local name of the repository to remove.

    namespace
        (string) The namespace scope for this request.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.repo_remove NAME

    """
    if namespace:
        if kvflags:
            kvflags.update({"namespace": namespace})
        else:
            kvflags = {"namespace": namespace}
    return _exec_true_return(
        commands=["repo", "remove", name], flags=flags, kvflags=kvflags
    )


def repo_update(namespace=None, flags=None, kvflags=None):
    """
    Update all charts repository.
    Return True if succeed, else the error message.

    namespace
        (string) The namespace scope for this request.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.repo_update

    """
    if namespace:
        if kvflags:
            kvflags.update({"namespace": namespace})
        else:
            kvflags = {"namespace": namespace}
    return _exec_true_return(commands=["repo", "update"], flags=flags, kvflags=kvflags)


def repo_manage(
    present=None, absent=None, prune=False, namespace=None, flags=None, kvflags=None
):
    """
    Manage charts repository.
    Return the summery of all actions.

    present
        (list) List of repository to be present. It's a list of dict: [{'name': 'local_name', 'url': 'repository_url'}]

    absent
        (list) List of local name repository to be absent.

    prune
        (boolean - default: False) If True, all repository already present but not in the present list would be removed.

    namespace
        (string) The namespace scope for this request.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.repo_manage present="[{'name': 'LOCAL_NAME', 'url': 'REPO_URL'}]" absent="['LOCAL_NAME']"

    """
    if present is None:
        present = []
    else:
        present = copy.deepcopy(present)
    if absent is None:
        absent = []
    else:
        absent = copy.deepcopy(absent)
    if namespace:
        if kvflags:
            kvflags.update({"namespace": namespace})
        else:
            kvflags = {"namespace": namespace}

    repos_present = repo_list(namespace=namespace, flags=flags, kvflags=kvflags)
    if not isinstance(repos_present, list):
        repos_present = []
    result = {"present": [], "added": [], "absent": [], "removed": [], "failed": []}

    for repo in present:
        if not (
            isinstance(repo, dict) and "name" in repo.keys() and "url" in repo.keys()
        ):
            raise CommandExecutionError(
                "Parameter present have to be formatted like "
                "[{'name': '<myRepoName>', 'url': '<myRepoUrl>'}]"
            )

        already_present = False
        for index, repo_present in enumerate(repos_present):
            if repo.get("name") == repo_present.get("name") and repo.get(
                "url"
            ) == repo_present.get("url"):
                result["present"].append(repo)
                repos_present.pop(index)
                already_present = True
                break

        if not already_present:
            repo_add_status = repo_add(
                repo.get("name"),
                repo.get("url"),
                namespace=namespace,
                flags=flags,
                kvflags=kvflags,
            )
            if isinstance(repo_add_status, bool) and repo_add_status:
                result["added"].append(repo)
            else:
                result["failed"].append(repo)

    for repo in repos_present:
        if prune:
            absent.append(repo.get("name"))
        elif not repo.get("name") in absent:
            result["present"].append(repo)

    for name in absent:
        remove_status = repo_remove(name, namespace=namespace)
        if isinstance(remove_status, bool) and remove_status:
            result["removed"].append(name)
        else:
            result["absent"].append(name)

    return result


def rollback(release, revision, namespace=None, flags=None, kvflags=None):
    """
    Rolls back a release to a previous revision.
    To see release revision number, execute the history module.
    Return True if succeed, else the error message.

    release
        (string) The name of the release to managed.

    revision
        (string) The revision number to roll back to.

    namespace
        (string) The namespace scope for this request.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.rollback RELEASE REVISION

        # In dry-run mode.
        salt '*' helm.rollback RELEASE REVISION flags=['dry-run']

    """
    if namespace:
        if kvflags:
            kvflags.update({"namespace": namespace})
        else:
            kvflags = {"namespace": namespace}
    return _exec_true_return(
        commands=["rollback", release, revision], flags=flags, kvflags=kvflags
    )


def search_hub(keyword, flags=None, kvflags=None):
    """
    Search the Helm Hub or an instance of Monocular for Helm charts.
    Return the research result if succeed, else the error message.

    keyword
        (string) The keyword to search in the hub.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.search_hub KEYWORD

        # In YAML format
        salt '*' helm.search_hub KEYWORD kvflags="{'output': 'yaml'}"

    """
    return _exec_dict_return(
        commands=["search", "hub", keyword], flags=flags, kvflags=kvflags
    )


def search_repo(keyword, flags=None, kvflags=None):
    """
    Search reads through all of the repositories configured on the system, and looks for matches. Search of these
    repositories uses the metadata stored on the system.
    Return the research result if succeed, else the error message.

    keyword
        (string) The keyword to search in the repo.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.search_hub KEYWORD

        # In YAML format
        salt '*' helm.search_hub KEYWORD kvflags="{'output': 'yaml'}"

    """
    return _exec_dict_return(
        commands=["search", "repo", keyword], flags=flags, kvflags=kvflags
    )


def show_all(chart, flags=None, kvflags=None):
    """
    Inspects a chart (directory, file, or URL) and displays all its content (values.yaml, Charts.yaml, README).
    Return chart information if succeed, else the error message.

    chart
        (string) The chart to inspect.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.show_all CHART

    """
    return _exec_string_return(
        commands=["show", "all", chart], flags=flags, kvflags=kvflags
    )


def show_chart(chart, flags=None, kvflags=None):
    """
    Inspects a chart (directory, file, or URL) and displays the contents of the Charts.yaml file.
    Return chart information if succeed, else the error message.

    chart
        (string) The chart to inspect.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.show_chart CHART

    """
    return _exec_string_return(
        commands=["show", "chart", chart], flags=flags, kvflags=kvflags
    )


def show_readme(chart, flags=None, kvflags=None):
    """
    Inspects a chart (directory, file, or URL) and displays the contents of the README file.
    Return chart information if succeed, else the error message.

    chart
        (string) The chart to inspect.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.show_readme CHART

    """
    return _exec_string_return(
        commands=["show", "readme", chart], flags=flags, kvflags=kvflags
    )


def show_values(chart, flags=None, kvflags=None):
    """
    Inspects a chart (directory, file, or URL) and displays the contents of the values.yaml file.
    Return chart information if succeed, else the error message.

    chart
        (string) The chart to inspect.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.show_values CHART

    """
    return _exec_string_return(
        commands=["show", "values", chart], flags=flags, kvflags=kvflags
    )


def status(release, namespace=None, flags=None, kvflags=None):
    """
    Show the status of the release.
    Return the release status if succeed, else the error message.

    release
        (string) The release to status.

    namespace
        (string) The namespace scope for this request.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.status RELEASE

        # In YAML format
        salt '*' helm.status RELEASE kvflags="{'output': 'yaml'}"

    """
    if namespace:
        if kvflags:
            kvflags.update({"namespace": namespace})
        else:
            kvflags = {"namespace": namespace}
    return _exec_dict_return(commands=["status", release], flags=flags, kvflags=kvflags)


def template(
    name, chart, values=None, output_dir=None, set=None, flags=None, kvflags=None
):
    """
    Render chart templates locally and display the output.
    Return the chart renderer if succeed, else the error message.

    name
        (string) The template name.

    chart
        (string) The chart to template.

    values
        (string) Absolute path to the values.yaml file.

    output_dir
        (string) Absolute path to the output directory.

    set
        (string or list) Set a values on the command line.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.template NAME CHART

        # With values file.
        salt '*' helm.template NAME CHART values='/path/to/values.yaml' output_dir='path/to/output/dir'

    """
    if values:
        if kvflags:
            kvflags.update({"values": values})
        else:
            kvflags = {"values": values}
    if set:
        if kvflags:
            kvflags.update({"set": set})
        else:
            kvflags = {"set": set}
    if output_dir:
        kvflags.update({"output-dir": output_dir})
    return _exec_string_return(
        commands=["template", name, chart], flags=flags, kvflags=kvflags
    )


def test(release, flags=None, kvflags=None):
    """
    Runs the tests for a release.
    Return the test result if succeed, else the error message.

    release
        (string) The release name to test.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.test RELEASE

    """
    return _exec_string_return(commands=["test", release], flags=flags, kvflags=kvflags)


def uninstall(release, namespace=None, flags=None, kvflags=None):
    """
    Uninstall the release name.
    Return True if succeed, else the error message.

    release
        (string) The name of the release to managed.

    namespace
        (string) The namespace scope for this request.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.uninstall RELEASE

        # In dry-run mode.
        salt '*' helm.uninstall RELEASE flags=['dry-run']

    """
    if namespace:
        if kvflags:
            kvflags.update({"namespace": namespace})
        else:
            kvflags = {"namespace": namespace}
    return _exec_true_return(
        commands=["uninstall", release], flags=flags, kvflags=kvflags
    )


def upgrade(
    release,
    chart,
    values=None,
    version=None,
    namespace=None,
    set=None,
    flags=None,
    kvflags=None,
):
    """
    Upgrades a release to a new version of a chart.
    Return True if succeed, else the error message.

    release
        (string) The name of the release to managed.

    chart
        (string) The chart to managed.

    values
        (string) Absolute path to the values.yaml file.

    version
        (string) The exact chart version to install. If this is not specified, the latest version is installed.

    namespace
        (string) The namespace scope for this request.

    set
        (string or list) Set a values on the command line.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.upgrade RELEASE CHART

        # In dry-run mode.
        salt '*' helm.upgrade RELEASE CHART flags=['dry-run']

        # With values file.
        salt '*' helm.upgrade RELEASE CHART values='/path/to/values.yaml'

    """
    if values:
        if kvflags:
            kvflags.update({"values": values})
        else:
            kvflags = {"values": values}
    if version:
        if kvflags:
            kvflags.update({"version": version})
        else:
            kvflags = {"version": version}
    if namespace:
        if kvflags:
            kvflags.update({"namespace": namespace})
        else:
            kvflags = {"namespace": namespace}
    if set:
        if kvflags:
            kvflags.update({"set": set})
        else:
            kvflags = {"set": set}
    return _exec_true_return(
        commands=["upgrade", release, chart], flags=flags, kvflags=kvflags
    )


def verify(path, flags=None, kvflags=None):
    """
    Verify that the given chart has a valid provenance file.
    Return True if succeed, else the error message.

    path
        (string) The path to the chart file.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.verify PATH

    """
    return _exec_true_return(commands=["verify", path], flags=flags, kvflags=kvflags)


def version(flags=None, kvflags=None):
    """
    Show the version for Helm.
    Return version information if succeed, else the error message.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    CLI Example:

    .. code-block:: bash

        salt '*' helm.version

    """
    return _exec_string_return(commands=["version"], flags=flags, kvflags=kvflags)



"""
mac_utils tests
"""

import os
import plistlib
import subprocess
import xml.parsers.expat

import pytest

import salt.modules.cmdmod as cmd
import salt.utils.mac_utils as mac_utils
import salt.utils.platform
from salt.exceptions import CommandExecutionError, SaltInvocationError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, MockTimedProc, mock_open, patch
from tests.support.unit import TestCase


@pytest.mark.skip_unless_on_darwin
class MacUtilsTestCase(TestCase, LoaderModuleMockMixin):
    """
    test mac_utils salt utility
    """

    def setup_loader_modules(self):
        return {mac_utils: {}}

    def test_execute_return_success_not_supported(self):
        """
        test execute_return_success function
        command not supported
        """
        mock_cmd = MagicMock(
            return_value={"retcode": 0, "stdout": "not supported", "stderr": "error"}
        )
        with patch.object(mac_utils, "_run_all", mock_cmd):
            self.assertRaises(
                CommandExecutionError, mac_utils.execute_return_success, "dir c:\\"
            )

    def test_execute_return_success_command_failed(self):
        """
        test execute_return_success function
        command failed
        """
        mock_cmd = MagicMock(
            return_value={"retcode": 1, "stdout": "spongebob", "stderr": "error"}
        )
        with patch.object(mac_utils, "_run_all", mock_cmd):
            self.assertRaises(
                CommandExecutionError, mac_utils.execute_return_success, "dir c:\\"
            )

    def test_execute_return_success_command_succeeded(self):
        """
        test execute_return_success function
        command succeeded
        """
        mock_cmd = MagicMock(return_value={"retcode": 0, "stdout": "spongebob"})
        with patch.object(mac_utils, "_run_all", mock_cmd):
            ret = mac_utils.execute_return_success("dir c:\\")
            self.assertEqual(ret, True)

    def test_execute_return_result_command_failed(self):
        """
        test execute_return_result function
        command failed
        """
        mock_cmd = MagicMock(
            return_value={"retcode": 1, "stdout": "spongebob", "stderr": "squarepants"}
        )
        with patch.object(mac_utils, "_run_all", mock_cmd):
            self.assertRaises(
                CommandExecutionError, mac_utils.execute_return_result, "dir c:\\"
            )

    def test_execute_return_result_command_succeeded(self):
        """
        test execute_return_result function
        command succeeded
        """
        mock_cmd = MagicMock(return_value={"retcode": 0, "stdout": "spongebob"})
        with patch.object(mac_utils, "_run_all", mock_cmd):
            ret = mac_utils.execute_return_result("dir c:\\")
            self.assertEqual(ret, "spongebob")

    def test_parse_return_space(self):
        """
        test parse_return function
        space after colon
        """
        self.assertEqual(
            mac_utils.parse_return("spongebob: squarepants"), "squarepants"
        )

    def test_parse_return_new_line(self):
        """
        test parse_return function
        new line after colon
        """
        self.assertEqual(
            mac_utils.parse_return("spongebob:\nsquarepants"), "squarepants"
        )

    def test_parse_return_no_delimiter(self):
        """
        test parse_return function
        no delimiter
        """
        self.assertEqual(mac_utils.parse_return("squarepants"), "squarepants")

    def test_validate_enabled_on(self):
        """
        test validate_enabled function
        test on
        """
        self.assertEqual(mac_utils.validate_enabled("On"), "on")

    def test_validate_enabled_off(self):
        """
        test validate_enabled function
        test off
        """
        self.assertEqual(mac_utils.validate_enabled("Off"), "off")

    def test_validate_enabled_bad_string(self):
        """
        test validate_enabled function
        test bad string
        """
        self.assertRaises(SaltInvocationError, mac_utils.validate_enabled, "bad string")

    def test_validate_enabled_non_zero(self):
        """
        test validate_enabled function
        test non zero
        """
        for x in range(1, 179, 3):
            self.assertEqual(mac_utils.validate_enabled(x), "on")

    def test_validate_enabled_0(self):
        """
        test validate_enabled function
        test 0
        """
        self.assertEqual(mac_utils.validate_enabled(0), "off")

    def test_validate_enabled_true(self):
        """
        test validate_enabled function
        test True
        """
        self.assertEqual(mac_utils.validate_enabled(True), "on")

    def test_validate_enabled_false(self):
        """
        test validate_enabled function
        test False
        """
        self.assertEqual(mac_utils.validate_enabled(False), "off")

    def test_launchctl(self):
        """
        test launchctl function
        """
        mock_cmd = MagicMock(
            return_value={"retcode": 0, "stdout": "success", "stderr": "none"}
        )
        with patch("salt.utils.mac_utils.__salt__", {"cmd.run_all": mock_cmd}):
            ret = mac_utils.launchctl("enable", "org.salt.minion")
            self.assertEqual(ret, True)

    def test_launchctl_return_stdout(self):
        """
        test launchctl function and return stdout
        """
        mock_cmd = MagicMock(
            return_value={"retcode": 0, "stdout": "success", "stderr": "none"}
        )
        with patch("salt.utils.mac_utils.__salt__", {"cmd.run_all": mock_cmd}):
            ret = mac_utils.launchctl("enable", "org.salt.minion", return_stdout=True)
            self.assertEqual(ret, "success")

    def test_launchctl_error(self):
        """
        test launchctl function returning an error
        """
        mock_cmd = MagicMock(
            return_value={"retcode": 1, "stdout": "failure", "stderr": "test failure"}
        )
        error = (
            "Failed to enable service:\n"
            "stdout: failure\n"
            "stderr: test failure\n"
            "retcode: 1"
        )
        with patch("salt.utils.mac_utils.__salt__", {"cmd.run_all": mock_cmd}):
            try:
                mac_utils.launchctl("enable", "org.salt.minion")
            except CommandExecutionError as exc:
                self.assertEqual(exc.message, error)

    @patch("salt.utils.path.os_walk")
    @patch("os.path.exists")
    def test_available_services_result(self, mock_exists, mock_os_walk):
        """
        test available_services results are properly formed dicts.
        """
        results = {"/Library/LaunchAgents": ["com.apple.lla1.plist"]}
        mock_os_walk.side_effect = _get_walk_side_effects(results)
        mock_exists.return_value = True

        plists = [{"Label": "com.apple.lla1"}]
        ret = _run_available_services(plists)

        file_path = os.sep + os.path.join(
            "Library", "LaunchAgents", "com.apple.lla1.plist"
        )
        if salt.utils.platform.is_windows():
            file_path = "c:" + file_path

        expected = {
            "com.apple.lla1": {
                "file_name": "com.apple.lla1.plist",
                "file_path": file_path,
                "plist": plists[0],
            }
        }
        self.assertEqual(ret, expected)

    @patch("salt.utils.path.os_walk")
    @patch("os.path.exists")
    @patch("os.listdir")
    @patch("os.path.isdir")
    def test_available_services_dirs(
        self, mock_isdir, mock_listdir, mock_exists, mock_os_walk
    ):
        """
        test available_services checks all of the expected dirs.
        """
        results = {
            "/Library/LaunchAgents": ["com.apple.lla1.plist"],
            "/Library/LaunchDaemons": ["com.apple.lld1.plist"],
            "/System/Library/LaunchAgents": ["com.apple.slla1.plist"],
            "/System/Library/LaunchDaemons": ["com.apple.slld1.plist"],
            "/Users/saltymcsaltface/Library/LaunchAgents": ["com.apple.uslla1.plist"],
        }

        mock_os_walk.side_effect = _get_walk_side_effects(results)
        mock_listdir.return_value = ["saltymcsaltface"]
        mock_isdir.return_value = True
        mock_exists.return_value = True

        plists = [
            {"Label": "com.apple.lla1"},
            {"Label": "com.apple.lld1"},
            {"Label": "com.apple.slla1"},
            {"Label": "com.apple.slld1"},
            {"Label": "com.apple.uslla1"},
        ]
        ret = _run_available_services(plists)

        self.assertEqual(len(ret), 5)

    @patch("salt.utils.path.os_walk")
    @patch("os.path.exists")
    @patch("plistlib.load")
    def test_available_services_broken_symlink(
        self, mock_read_plist, mock_exists, mock_os_walk
    ):
        """
        test available_services when it encounters a broken symlink.
        """
        results = {
            "/Library/LaunchAgents": ["com.apple.lla1.plist", "com.apple.lla2.plist"]
        }
        mock_os_walk.side_effect = _get_walk_side_effects(results)
        mock_exists.side_effect = [True, False]

        plists = [{"Label": "com.apple.lla1"}]
        ret = _run_available_services(plists)

        file_path = os.sep + os.path.join(
            "Library", "LaunchAgents", "com.apple.lla1.plist"
        )
        if salt.utils.platform.is_windows():
            file_path = "c:" + file_path

        expected = {
            "com.apple.lla1": {
                "file_name": "com.apple.lla1.plist",
                "file_path": file_path,
                "plist": plists[0],
            }
        }
        self.assertEqual(ret, expected)

    @patch("salt.utils.path.os_walk")
    @patch("os.path.exists")
    @patch("salt.utils.mac_utils.__salt__")
    def test_available_services_binary_plist(
        self,
        mock_run,
        mock_exists,
        mock_os_walk,
    ):
        """
        test available_services handles binary plist files.
        """
        results = {"/Library/LaunchAgents": ["com.apple.lla1.plist"]}
        mock_os_walk.side_effect = _get_walk_side_effects(results)
        mock_exists.return_value = True

        plists = [{"Label": "com.apple.lla1"}]

        file_path = os.sep + os.path.join(
            "Library", "LaunchAgents", "com.apple.lla1.plist"
        )
        if salt.utils.platform.is_windows():
            file_path = "c:" + file_path

        ret = _run_available_services(plists)

        expected = {
            "com.apple.lla1": {
                "file_name": "com.apple.lla1.plist",
                "file_path": file_path,
                "plist": plists[0],
            }
        }
        self.assertEqual(ret, expected)

    @patch("salt.utils.path.os_walk")
    @patch("os.path.exists")
    def test_available_services_invalid_file(self, mock_exists, mock_os_walk):
        """
        test available_services excludes invalid files.
        The py3 plistlib raises an InvalidFileException when a plist
        file cannot be parsed.
        """
        results = {"/Library/LaunchAgents": ["com.apple.lla1.plist"]}
        mock_os_walk.side_effect = _get_walk_side_effects(results)
        mock_exists.return_value = True

        plists = [{"Label": "com.apple.lla1"}]

        mock_load = MagicMock()
        mock_load.side_effect = plistlib.InvalidFileException
        with patch("salt.utils.files.fopen", mock_open()):
            with patch("plistlib.load", mock_load):
                ret = mac_utils._available_services()

        self.assertEqual(len(ret), 0)

    @patch("salt.utils.mac_utils.__salt__")
    @patch("salt.utils.path.os_walk")
    @patch("os.path.exists")
    def test_available_services_expat_error(self, mock_exists, mock_os_walk, mock_run):
        """
        test available_services excludes files with expat errors.

        Poorly formed XML will raise an ExpatError on py2. It will
        also be raised by some almost-correct XML on py3.
        """
        results = {"/Library/LaunchAgents": ["com.apple.lla1.plist"]}
        mock_os_walk.side_effect = _get_walk_side_effects(results)
        mock_exists.return_value = True

        file_path = os.sep + os.path.join(
            "Library", "LaunchAgents", "com.apple.lla1.plist"
        )
        if salt.utils.platform.is_windows():
            file_path = "c:" + file_path

        mock_load = MagicMock()
        mock_load.side_effect = xml.parsers.expat.ExpatError
        with patch("salt.utils.files.fopen", mock_open()):
            with patch("plistlib.load", mock_load):
                ret = mac_utils._available_services()

        self.assertEqual(len(ret), 0)

    @patch("salt.utils.mac_utils.__salt__")
    @patch("salt.utils.path.os_walk")
    @patch("os.path.exists")
    def test_available_services_value_error(self, mock_exists, mock_os_walk, mock_run):
        """
        test available_services excludes files with ValueErrors.
        """
        results = {"/Library/LaunchAgents": ["com.apple.lla1.plist"]}
        mock_os_walk.side_effect = _get_walk_side_effects(results)
        mock_exists.return_value = True

        file_path = os.sep + os.path.join(
            "Library", "LaunchAgents", "com.apple.lla1.plist"
        )
        if salt.utils.platform.is_windows():
            file_path = "c:" + file_path

        mock_load = MagicMock()
        mock_load.side_effect = ValueError
        with patch("salt.utils.files.fopen", mock_open()):
            with patch("plistlib.load", mock_load):
                ret = mac_utils._available_services()

        self.assertEqual(len(ret), 0)

    def test_bootout_retcode_36_success(self):
        """
        Make sure that if we run a `launchctl bootout` cmd and it returns
        36 that we treat it as a success.
        """
        proc = MagicMock(
            return_value=MockTimedProc(stdout=None, stderr=None, returncode=36)
        )
        with patch("salt.utils.timed_subprocess.TimedProc", proc):
            with patch(
                "salt.utils.mac_utils.__salt__", {"cmd.run_all": cmd._run_all_quiet}
            ):
                ret = mac_utils.launchctl("bootout", "org.salt.minion")
        self.assertEqual(ret, True)

    def test_bootout_retcode_99_fail(self):
        """
        Make sure that if we run a `launchctl bootout` cmd and it returns
        something other than 0 or 36 that we treat it as a fail.
        """
        error = (
            "Failed to bootout service:\n"
            "stdout: failure\n"
            "stderr: test failure\n"
            "retcode: 99"
        )
        proc = MagicMock(
            return_value=MockTimedProc(
                stdout=b"failure", stderr=b"test failure", returncode=99
            )
        )
        with patch("salt.utils.timed_subprocess.TimedProc", proc):
            with patch(
                "salt.utils.mac_utils.__salt__", {"cmd.run_all": cmd._run_all_quiet}
            ):
                try:
                    mac_utils.launchctl("bootout", "org.salt.minion")
                except CommandExecutionError as exc:
                    self.assertEqual(exc.message, error)

    def test_not_bootout_retcode_36_fail(self):
        """
        Make sure that if we get a retcode 36 on non bootout cmds
        that we still get a failure.
        """
        error = (
            "Failed to bootstrap service:\n"
            "stdout: failure\n"
            "stderr: test failure\n"
            "retcode: 36"
        )
        proc = MagicMock(
            return_value=MockTimedProc(
                stdout=b"failure", stderr=b"test failure", returncode=36
            )
        )
        with patch("salt.utils.timed_subprocess.TimedProc", proc):
            with patch(
                "salt.utils.mac_utils.__salt__", {"cmd.run_all": cmd._run_all_quiet}
            ):
                try:
                    mac_utils.launchctl("bootstrap", "org.salt.minion")
                except CommandExecutionError as exc:
                    self.assertEqual(exc.message, error)

    def test_git_is_stub(self):
        mock_check_call = MagicMock(
            side_effect=subprocess.CalledProcessError(cmd="", returncode=2)
        )
        with patch("salt.utils.mac_utils.subprocess.check_call", mock_check_call):
            self.assertEqual(mac_utils.git_is_stub(), True)

    @patch("salt.utils.mac_utils.subprocess.check_call")
    def test_git_is_not_stub(self, mock_check_call):
        self.assertEqual(mac_utils.git_is_stub(), False)


def _get_walk_side_effects(results):
    """
    Data generation helper function for service tests.
    """

    def walk_side_effect(*args, **kwargs):
        return [(args[0], [], results.get(args[0], []))]

    return walk_side_effect


def _run_available_services(plists):
    mock_load = MagicMock()
    mock_load.side_effect = plists
    with patch("salt.utils.files.fopen", mock_open()):
        with patch("plistlib.load", mock_load):
            ret = mac_utils._available_services()
    return ret



"""
Monit state
===========

Manage monit states

.. code-block:: yaml

    monit_enable_service_monitoring:
      monit.monitor:
        - name: service

    monit_disable_service_monitoring:
      monit.unmonitor:
        - name: service

.. note::
    Use of these states require that the :mod:`monit <salt.modules.monit>`
    execution module is available.
"""


def __virtual__():
    """
    Only make this state available if the monit module is available.
    """
    if "monit.summary" in __salt__:
        return "monit"
    return (False, "monit module could not be loaded")


def monitor(name):
    """
    Get the summary from module monit and try to see if service is
    being monitored. If not then monitor the service.
    """
    ret = {"result": None, "name": name, "comment": "", "changes": {}}
    result = __salt__["monit.summary"](name)

    try:
        for key, value in result.items():
            if "Running" in value[name]:
                ret["comment"] = f"{name} is being being monitored."
                ret["result"] = True
            else:
                if __opts__["test"]:
                    ret["comment"] = f"Service {name} is set to be monitored."
                    ret["result"] = None
                    return ret
                __salt__["monit.monitor"](name)
                ret["comment"] = f"{name} started to be monitored."
                ret["changes"][name] = "Running"
                ret["result"] = True
                break
    except KeyError:
        ret["comment"] = f"{name} not found in configuration."
        ret["result"] = False

    return ret


def unmonitor(name):
    """
    Get the summary from module monit and try to see if service is
    being monitored. If it is then stop monitoring the service.
    """
    ret = {"result": None, "name": name, "comment": "", "changes": {}}
    result = __salt__["monit.summary"](name)

    try:
        for key, value in result.items():
            if "Not monitored" in value[name]:
                ret["comment"] = f"{name} is not being monitored."
                ret["result"] = True
            else:
                if __opts__["test"]:
                    ret["comment"] = f"Service {name} is set to be unmonitored."
                    ret["result"] = None
                    return ret
                __salt__["monit.unmonitor"](name)
                ret["comment"] = f"{name} stopped being monitored."
                ret["changes"][name] = "Not monitored"
                ret["result"] = True
                break
    except KeyError:
        ret["comment"] = f"{name} not found in configuration."
        ret["result"] = False

    return ret



import pytest

import salt.states.win_servermanager as win_servermanager
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {win_servermanager: {}}


def test_installed():
    """
    Test to install the windows feature
    """
    mock_list = MagicMock(
        side_effect=[
            {"spongebob": "squarepants"},
            {"squidward": "patrick"},
            {"spongebob": "squarepants"},
            {"spongebob": "squarepants", "squidward": "patrick"},
        ]
    )
    mock_install = MagicMock(
        return_value={
            "Success": True,
            "Restarted": False,
            "RestartNeeded": False,
            "ExitCode": 1234,
            "Features": {
                "squidward": {
                    "DisplayName": "Squidward",
                    "Message": "",
                    "RestartNeeded": True,
                    "SkipReason": 0,
                    "Success": True,
                }
            },
        }
    )
    with patch.dict(
        win_servermanager.__salt__,
        {
            "win_servermanager.list_installed": mock_list,
            "win_servermanager.install": mock_install,
        },
    ):
        ret = {
            "name": "spongebob",
            "changes": {},
            "result": True,
            "comment": "The following features are already installed:\n- spongebob",
        }
        assert win_servermanager.installed("spongebob") == ret

        with patch.dict(win_servermanager.__opts__, {"test": True}):
            ret = {
                "name": "spongebob",
                "result": None,
                "comment": "",
                "changes": {"spongebob": "Will be installed recurse=False"},
            }
            assert win_servermanager.installed("spongebob") == ret

        with patch.dict(win_servermanager.__opts__, {"test": False}):
            ret = {
                "name": "squidward",
                "result": True,
                "comment": "Installed the following:\n- squidward",
                "changes": {"squidward": {"new": "patrick", "old": ""}},
            }
            assert win_servermanager.installed("squidward") == ret


def test_removed():
    """
    Test to remove the windows feature
    """
    mock_list = MagicMock(
        side_effect=[
            {"spongebob": "squarepants"},
            {"squidward": "patrick"},
            {"spongebob": "squarepants", "squidward": "patrick"},
            {"spongebob": "squarepants"},
        ]
    )
    mock_remove = MagicMock(
        return_value={
            "Success": True,
            "RestartNeeded": False,
            "Restarted": False,
            "ExitCode": 1234,
            "Features": {
                "squidward": {
                    "DisplayName": "Squidward",
                    "Message": "",
                    "RestartNeeded": True,
                    "SkipReason": 0,
                    "Success": True,
                }
            },
        }
    )
    with patch.dict(
        win_servermanager.__salt__,
        {
            "win_servermanager.list_installed": mock_list,
            "win_servermanager.remove": mock_remove,
        },
    ):
        ret = {
            "name": "squidward",
            "changes": {},
            "result": True,
            "comment": "The following features are not installed:\n- squidward",
        }
        assert win_servermanager.removed("squidward") == ret

        with patch.dict(win_servermanager.__opts__, {"test": True}):
            ret = {
                "name": "squidward",
                "result": None,
                "comment": "",
                "changes": {"squidward": "Will be removed"},
            }
            assert win_servermanager.removed("squidward") == ret

        with patch.dict(win_servermanager.__opts__, {"test": False}):
            ret = {
                "name": "squidward",
                "result": True,
                "comment": "Removed the following:\n- squidward",
                "changes": {"squidward": {"new": "", "old": "patrick"}},
            }
            assert win_servermanager.removed("squidward") == ret



"""
    tests.support.pytest.loader
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Salt's Loader PyTest Mock Support
"""

import logging
import sys
import types
from collections import deque

import attr

from tests.support.mock import patch

log = logging.getLogger(__name__)


@attr.s(init=True, slots=True, frozen=True)
class LoaderModuleMock:

    setup_loader_modules = attr.ib(init=True)
    # These dunders should always exist at the module global scope
    salt_module_dunders = attr.ib(
        init=True,
        repr=False,
        kw_only=True,
        default=(
            "__opts__",
            "__salt__",
            "__runner__",
            "__context__",
            "__utils__",
            "__ext_pillar__",
            "__thorium__",
            "__states__",
            "__serializers__",
            "__ret__",
            "__grains__",
            "__pillar__",
            "__sdb__",
        ),
    )
    # These dunders might exist at the module global scope
    salt_module_dunders_optional = attr.ib(
        init=True,
        repr=False,
        kw_only=True,
        default=("__proxy__",),
    )
    # These dunders might exist at the function global scope
    salt_module_dunder_attributes = attr.ib(
        init=True,
        repr=False,
        kw_only=True,
        default=(
            # Salt states attributes
            "__env__",
            "__low__",
            "__instance_id__",
            "__orchestration_jid__",
            # Salt runners attributes
            "__jid_event__",
            # Salt cloud attributes
            "__active_provider_name__",
            # Proxy Minions
            "__proxyenabled__",
        ),
    )
    _finalizers = attr.ib(
        init=False, repr=False, hash=False, default=attr.Factory(deque)
    )

    def start(self):
        module_globals = {dunder: {} for dunder in self.salt_module_dunders}
        for module, globals_to_mock in self.setup_loader_modules.items():
            log.trace(
                "Setting up loader globals for %s; globals: %s", module, globals_to_mock
            )
            if not isinstance(module, types.ModuleType):
                raise RuntimeError(
                    "The dictionary keys returned by setup_loader_modules() "
                    "must be an imported module, not {}".format(type(module))
                )
            if not isinstance(globals_to_mock, dict):
                raise RuntimeError(
                    "The dictionary values returned by setup_loader_modules() "
                    "must be a dictionary, not {}".format(type(globals_to_mock))
                )
            for key in self.salt_module_dunders:
                if not hasattr(module, key):
                    # Set the dunder name as an attribute on the module if not present
                    setattr(module, key, {})
                    # Remove the added attribute after the test finishes
                    self.addfinalizer(delattr, module, key)

            # Patch sys.modules as the first step
            self._patch_sys_modules(globals_to_mock)

            # Now patch the module globals
            # We actually want to grab a copy of the module globals so that if mocking
            # multiple modules, and at least one of the modules has a function to path,
            # the patch only happens on the module it's supposed to patch and not all of them.
            # It's not a deepcopy because we want to maintain the reference to the salt dunders
            # added in the start of this function
            self._patch_module_globals(module, globals_to_mock, module_globals.copy())

    def stop(self):
        while self._finalizers:
            func, args, kwargs = self._finalizers.popleft()
            func_repr = self._format_callback(func, args, kwargs)
            try:
                log.trace("Calling finalizer %s", func_repr)
                func(*args, **kwargs)
            except Exception as exc:  # pylint: disable=broad-except
                log.error(
                    "Failed to run finalizer %s: %s",
                    func_repr,
                    exc,
                    exc_info=True,
                )

    def addfinalizer(self, func, *args, **kwargs):
        """
        Register a function to run when stopping
        """
        self._finalizers.append((func, args, kwargs))

    def _format_callback(self, callback, args, kwargs):
        callback_str = f"{callback.__qualname__}("
        if args:
            callback_str += ", ".join([repr(arg) for arg in args])
        if kwargs:
            callback_str += ", ".join([f"{k}={v!r}" for (k, v) in kwargs.items()])
        callback_str += ")"
        return callback_str

    def _patch_sys_modules(self, mocks):
        if "sys.modules" not in mocks:
            return
        sys_modules = mocks["sys.modules"]
        if not isinstance(sys_modules, dict):
            raise RuntimeError(
                f"'sys.modules' must be a dictionary not: {type(sys_modules)}"
            )
        patcher = patch.dict(sys.modules, values=sys_modules)
        patcher.start()
        self.addfinalizer(patcher.stop)

    def _patch_module_globals(self, module, mocks, module_globals):
        salt_dunder_dicts = self.salt_module_dunders + self.salt_module_dunders_optional
        allowed_salt_dunders = salt_dunder_dicts + self.salt_module_dunder_attributes
        for key in mocks:
            if key == "sys.modules":
                # sys.modules is addressed on another function
                continue

            if key.startswith("__"):
                if key in ("__init__", "__virtual__"):
                    raise RuntimeError(
                        "No need to patch {!r}. Passed loader module dict: {}".format(
                            key,
                            self.setup_loader_modules,
                        )
                    )
                elif key not in allowed_salt_dunders:
                    raise RuntimeError(
                        "Don't know how to handle {!r}. Passed loader module dict: {}".format(
                            key,
                            self.setup_loader_modules,
                        )
                    )
                elif key in salt_dunder_dicts and not hasattr(module, key):
                    # Add the key as a dictionary attribute to the module so it can be patched by `patch.dict`'
                    setattr(module, key, {})
                    # Remove the added attribute after the test finishes
                    self.addfinalizer(delattr, module, key)

            if not hasattr(module, key):
                # Set the key as an attribute so it can be patched
                setattr(module, key, None)
                # Remove the added attribute after the test finishes
                self.addfinalizer(delattr, module, key)
            module_globals[key] = mocks[key]

        # Patch the module!
        log.trace("Patching globals for %s; globals: %s", module, module_globals)
        patcher = patch.multiple(module, **module_globals)
        patcher.start()
        self.addfinalizer(patcher.stop)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.stop()



"""
Allow for flow based timers. These timers allow for a sleep to exist across
multiple runs of the flow
"""

import time


def hold(name, seconds):
    """
    Wait for a given period of time, then fire a result of True, requiring
    this state allows for an action to be blocked for evaluation based on
    time

    USAGE:

    .. code-block:: yaml

        hold_on_a_moment:
          timer.hold:
            - seconds: 30
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}
    start = time.time()
    if "timer" not in __context__:
        __context__["timer"] = {}
    if name not in __context__["timer"]:
        __context__["timer"][name] = start
    if (start - __context__["timer"][name]) > seconds:
        ret["result"] = True
        __context__["timer"][name] = start
    return ret



"""
Use hiera data as a Pillar source
"""

import logging

import salt.utils.path
import salt.utils.yaml

# Set up logging
log = logging.getLogger(__name__)


def __virtual__():
    """
    Only return if hiera is installed
    """
    return "hiera" if salt.utils.path.which("hiera") else False


def ext_pillar(
    minion_id, pillar, conf  # pylint: disable=W0613  # pylint: disable=W0613
):
    """
    Execute hiera and return the data
    """
    cmd = f"hiera -c {conf}"
    for key, val in __grains__.items():
        if isinstance(val, str):
            cmd += f" {key}='{val}'"
    try:
        data = salt.utils.yaml.safe_load(__salt__["cmd.run"](cmd))
    except Exception:  # pylint: disable=broad-except
        log.critical("Hiera YAML data failed to parse from conf %s", conf)
        return {}
    return data



"""
Verify salt-ssh fails with a retcode > 0 when a highstate verification fails.
``state.show_highstate`` does not validate this.
"""

import pytest

from salt.defaults.exitcodes import EX_AGGREGATE

pytestmark = [
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
    pytest.mark.slow_test,
]


@pytest.fixture(scope="module", autouse=True)
def state_tree_req_fail(base_env_state_tree_root_dir):
    top_file = """
    base:
      'localhost':
        - fail_req
      '127.0.0.1':
        - fail_req
    """
    state_file = """
    This has an invalid requisite:
      test.nop:
        - name: foo
        - require_in:
          - file.managed: invalid_requisite
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_state_tree_root_dir
    )
    state_tempfile = pytest.helpers.temp_file(
        "fail_req.sls", state_file, base_env_state_tree_root_dir
    )
    with top_tempfile, state_tempfile:
        yield


@pytest.mark.parametrize(
    "args,retcode",
    (
        (("state.sls", "fail_req"), EX_AGGREGATE),
        (("state.highstate",), EX_AGGREGATE),
        (("state.show_sls", "fail_req"), EX_AGGREGATE),
        (("state.show_low_sls", "fail_req"), EX_AGGREGATE),
        # state.show_lowstate exits with 0 for non-ssh as well
        (("state.show_lowstate",), 0),
        (("state.top", "top.sls"), EX_AGGREGATE),
    ),
)
def test_it(salt_ssh_cli, args, retcode):
    ret = salt_ssh_cli.run(*args)
    assert ret.returncode == retcode
    assert isinstance(ret.data, list)
    assert ret.data
    assert isinstance(ret.data[0], str)
    assert ret.data[0].startswith(
        "Invalid requisite in require: file.managed for invalid_requisite"
    )



"""
Validate the mac-defaults module
"""

import pytest

from tests.support.case import ModuleCase

DEFAULT_DOMAIN = "com.apple.AppleMultitouchMouse"
DEFAULT_KEY = "MouseHorizontalScroll"
DEFAULT_VALUE = "0"


@pytest.mark.destructive_test
@pytest.mark.skip_if_not_root
@pytest.mark.skip_unless_on_darwin
class MacDefaultsModuleTest(ModuleCase):
    """
    Integration tests for the mac_default module
    """

    def test_macdefaults_write_read(self):
        """
        Tests that writes and reads macdefaults
        """
        write_domain = self.run_function(
            "macdefaults.write", [DEFAULT_DOMAIN, DEFAULT_KEY, DEFAULT_VALUE]
        )
        self.assertTrue(write_domain)

        read_domain = self.run_function(
            "macdefaults.read", [DEFAULT_DOMAIN, DEFAULT_KEY]
        )
        self.assertTrue(read_domain)
        self.assertEqual(read_domain, DEFAULT_VALUE)
